"""
Functions wrap PiCloud calls

"""
"""
Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""

"""TODO:
Add rest publishing..
result (to get things from rest)
"""

import os
import random
import sys
import getpass
import time
try:
    import json
except:
    # Python 2.5 compatibility
    import simplejson as json

from itertools import izip
from subprocess import Popen, PIPE

# TODO: Must move this out of CLI into shell-exec.....
from .. import shell
from .. import bucket
from .. import _getscicloud
from ..scicloud import CloudException
from ..util import template, OrderedDict
from ..util.xrange_helper import PiecewiseXrange, maybe_xrange_iter
from ..rest import _invoke

import logging
scicloudLog = logging.getLogger('Cloud.functions')

def _gen_shell_kwargs(extra_args):
    kwargs = {}
    for arg, val in extra_args.items():
        if val == None: #default/not present
            continue        
        if arg == 'depends_on':
            val = parse_jids(val)
        kwargs['_' + arg] = val
    return kwargs
        

def execute(command, args, return_file, ignore_exit_status, cwd=None, **kwargs):
    command = ' '.join(command)
    
    #return scicloud.call(execute_program, command)
    arg_dct = template.extract_args(args, allow_map = False)
    kwargs = _gen_shell_kwargs(kwargs)
    
    jid = shell.execute(command, arg_dct, return_file, ignore_exit_status, cwd, **kwargs)
    
    scicloud = _getscicloud()
    if scicloud.adapter.connection.is_simulated(): # on simulation wait for job to complete
        return result(str(jid))
    else:  # only return jid when not simulating
        return jid
    
def execute_map(command, args, mapargs, argfiles, duplicates,                
                return_file, ignore_exit_status, cwd=None, **kwargs):
    command = ' '.join(command)
    arg_dct = template.extract_args(args, allow_map = False)
    if not mapargs and not argfiles:
        if duplicates <= 1:
            raise ValueError('At least one -n or -N must be provided')
    
    maparg_dct = {}
    if mapargs:
        maparg_dct = template.extract_args(mapargs, allow_map = True)
    if argfiles:  # get all param=filenames
        maparg_file_dct = template.extract_args(argfiles, allow_map=False)
        for param, filename in maparg_file_dct.items():
            dir, rel_filename = os.path.split(filename)
            
            if param in maparg_dct:
                raise ValueError('key %s cannot be defined in both -a and -n' % param)
            with open(filename) as f:
                lines = [s.strip() for s in f.readlines()]                                
                final_args = []
                for line in lines:
                    if not line: # ignore blanks
                        continue
                    if line.startswith('@'): #see shell._handle_args_upload. need to make this releative
                        line = '@' + dir + '/' + line[1:]
                    final_args.append(line)
                maparg_dct[param] = final_args
            
        
    kwargs = _gen_shell_kwargs(kwargs)
    
    if duplicates > 1:
        if not maparg_dct:
            maparg_dct = {'' : ['']*duplicates}
        else:
            new_map_arg_dct = {}
            for key, vals in maparg_dct.items():
                new_map_arg_dct[key] = vals*duplicates
            maparg_dct=new_map_arg_dct

    
    jids = shell.execute_map(command, arg_dct, maparg_dct, return_file, 
                             ignore_exit_status, cwd, **kwargs)
    scicloud = _getscicloud()
    if scicloud.adapter.connection.is_simulated(): # on simulation wait for job to complete
        return result('%s-%s' % (jids[0], jids[-1]))
    else:  # only return jid when not simulating
        if isinstance(jids, xrange):
            return '%s-%s' % (jids[0], jids[-1])
        else:
            return jids
        
def rest_publish(command, label, return_file, ignore_exit_status, **kwargs):
    command = ' '.join(command)

    kwargs = _gen_shell_kwargs(kwargs)
    
    return shell.rest_publish(command, label, return_file, ignore_exit_status, **kwargs)

def cron_register(command, label, schedule, return_file, ignore_exit_status, **kwargs):
    command = ' '.join(command)

    kwargs = _gen_shell_kwargs(kwargs)
    
    return shell.cron_register(command, label, schedule, return_file, ignore_exit_status, **kwargs)

"""other rest"""
def _handle_file_upload(arg):
    if arg.startswith('@'):
        filename = arg[1:]
        return open(filename,'rb')
    else:
        return arg
    

def rest_invoke(label, args):
    arg_dct = template.extract_args(args, allow_map = False)
    listified_args = dict((name, [_handle_file_upload(val)]) for name, val in arg_dct.items())
    return _invoke(label, listified_args, is_map = False)

def rest_invoke_map(label, args, mapargs):
    if not mapargs:
        raise ValueError('At least one -n (--map-data) parameter must be provided')
    
    single_arg_dct = template.extract_args(args, allow_map = False)
    map_arg_dct = template.extract_args(mapargs, allow_map = True)
    
    for kwd, arg in single_arg_dct.items():
        map_arg_dct[kwd] = [arg]
        
    for kwd, arglist in map_arg_dct.items():
        map_arg_dct[kwd] = [_handle_file_upload(arg) for arg in arglist]
    
    jids = _invoke(label, map_arg_dct, is_map = True)
    if isinstance(jids, xrange):
        return '%s-%s' % (jids[0], jids[-1])
    else:
        return jids

def bucket_make_public(obj_path, prefix, header_args, reset_headers):
    headers = template.extract_args(header_args, allow_map=False)
    return bucket.make_public(obj_path, prefix, headers, reset_headers)
    
""" todo: kill, status, etc. on formatted jids"""
def status(jids):
    jids = parse_jids(jids)
    scicloud = _getscicloud()
    statuses = scicloud.status(jids)
    # encode answers with dct
    out_dct = OrderedDict(((jid, status) for jid, status in izip(jids.my_iter(), statuses)))
    return out_dct

def join(jids, timeout=None):
    jids = parse_jids(jids)
    scicloud = _getscicloud()
    return scicloud.join(jids,timeout)    

class _JSON_String(str):
    json_encoded = True
    
    def __init__(self, contents):
        super(_JSON_String, self).__init__(contents)

def result(jids, timeout=None, ignore_errors=False): 
    jids = parse_jids(jids)
    scicloud = _getscicloud()
    results = scicloud.result(jids, timeout, ignore_errors)
    result_dct = OrderedDict()
    for jid, result in izip(maybe_xrange_iter(jids), results):
        if isinstance(result, CloudException):
            result = 'CloudException: ' + str(result)
        if isinstance(result, basestring):
            pass
        elif isinstance(result, (int, bool, long, float)):
            pass # keep as primitive
        else:
            try:
                result = json.dumps(result)
            except (TypeError, UnicodeDecodeError):
                raise ValueError('Cannot view result of jid %s. It cannot be represented in JSON.' % jid)
            else:
                result = _JSON_String(result)
        result_dct[jid] = result
    return result_dct    

def info(jids, info_requested=None):
    jids = parse_jids(jids)
    if info_requested:
        info_requested = info_requested.split(',')
    scicloud = _getscicloud()
    return scicloud.info(jids, info_requested)    

def kill(jids):
    jids = parse_jids(jids)
    scicloud = _getscicloud()
    return scicloud.kill(jids)    

def delete(jids):
    jids = parse_jids(jids)
    scicloud = _getscicloud()
    return scicloud.delete(jids)    

def parse_jids(jid_str):
    """Parse jid a-b,c format (same as used for rest"""        
        
    ujids = PiecewiseXrange() 
    try:
        for jid in jid_str.split(','):
            if '-' in jid:
                min_jid, max_jid = jid.split('-')
                ujids.append(xrange(int(min_jid), int(max_jid)+1))
            else:
                ujids.append(int(jid))
    except ValueError:
        raise ValueError("jids must be integers sperated by '-' or ','")
    ujids.to_xrange_mode()
    return ujids

"""Other"""
def ssh(jid, timeout=None, cmd=None):
    """SSH into a job"""
    from ..shortcuts.ssh import get_ssh_info
    from ..util.common import ssh_session, fix_path

    ssh_info = get_ssh_info(jid, timeout)
    username = ssh_info['username']
    hostname = ssh_info['address']
    key_path = fix_path(ssh_info['identity'])
    port = ssh_info['port']
    
    status, stdout, stderr = ssh_session(username, hostname, key_path, port, run_cmd=cmd)
    if status:
        if stdout:
            sys.stdout.write(stdout)
        if stderr:
            sys.stderr.write(stderr)        
        sys.exit(status)
        
    if cmd and stdout:
        return stdout

def exec_shell(timeout=None, keep_alive=False, **kwargs):
    from ..util.common import ssh_server_job
    
    kwargs = _gen_shell_kwargs(kwargs)
    kwargs['_label'] = 'exec-shell'
    kwargs['_priority'] = 1 
    kwargs['_restartable'] = False
        
    scicloud, params = shell._get_scicloud_and_params('', kwargs)
    params['func_name'] = '<ssh session>'
    
    jid = scicloud.adapter.job_call(params, ssh_server_job, (keep_alive,), {})
    
    print 'Job requested as jid %s. SSHing in..' % jid
        
    try:
        ssh(jid, timeout)    
    except Exception:
        scicloud.kill(jid)
        raise




"""
Cloudlog controls the logging of all scicloud related messages

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import os
import sys
import logging


from . import scicloudconfig as cc
from .util import fix_sudo_path

c = \
"""Filename where scicloud log messages should be written.
This path is relative to ~/.piscicloud/"""
logFilename = cc.logging_configurable('log_filename',
                                     default='scicloud.log',  #NOTE: will not create directories
                                     comment =c)
c = \
"""Should log_filename (default of scicloud.log) be written with scicloud log messages?
Note that scicloud will always emit logging messages; this option controls if scicloud should have its own log."""
saveLog = cc.logging_configurable('save_log',
                                 default=True,
                                 comment=c)
c = \
"""logging level for scicloud messages.
This affects both messages saved to the scicloud log file and sent through the python logging system.
See http://docs.python.org/library/logging.html for more information"""
logLevel = cc.logging_configurable('log_level',
                                  default=logging.getLevelName(logging.DEBUG),
                                  comment=c)

c = \
"""logging level for printing scicloud log messages to console.
Must be equal or higher than log_level"""
printLogLevel = cc.logging_configurable('print_log_level',
                                  default=logging.getLevelName(logging.ERROR),
                                  comment=c)

c = """Number of days to wait until purging old serialization log directories"""
purgeDays = cc.logging_configurable('purge_days',
                                        default=7,
                                        comment =c)

by_pid_dir = 'scicloudlog-by-pid'

datefmt = '%a %b %d %H:%M:%S %Y'
logfmt_main = "[%(asctime)s] - [%(levelname)s] - %(name)s: %(message)s"
logfmt_pid = "[%(asctime)s] - [%(levelname)s] - %(name)s(%(process)d): %(message)s"

class NullHandler(logging.Handler):
    """A handler that does nothing"""
    def emit(self, record):
        pass


def purge_old_logs(pid_log_dir, mylog):
    """Remove contents in by-pid purgedays
    Warning: Potential errors if log file has been opened > purgedays and not written to
    Highly unlikely"""
    import errno, stat, time
    
    try:
        logs = os.listdir(pid_log_dir)
    except OSError, e:
        if e.errno != errno.ENOENT:
            mylog.debug('Could not purge %s due to %s', pid_log_dir, str(e))
        return
    
    now = time.time()
    allowed_difference = purgeDays * 24 * 3600 #purge days in seconds

    for log in logs:
        try:
            log_path = os.path.join(pid_log_dir, log)
            stat_result = os.stat(log_path)
        except OSError:
            mylog.warn('Could not stat %s', log, exc_info = True)
            continue        
        if  (now - stat_result.st_ctime) > allowed_difference:
            mylog.debug('Deleting %s (%s days old)', log, (now - stat_result.st_ctime)/(24*3600))
            try:
                os.remove(log_path)
            except OSError:
                mylog.warn('Could not delete %s', log_path, exc_info = True)            
    
    
def init_pid_log(base_path, mylog):
    from .util.cloghandler.cloghandler import ConcurrentRotatingFileHandler
    from warnings import warn
    import errno
        
    my_pid = str(os.getpid())      
    path = os.path.join(base_path, by_pid_dir)  

    try:
        os.makedirs(path)
    except OSError, e: #allowed to exist already
        if e.errno != errno.EEXIST:                
            warn('PiCloud cannot create directory %s. Error is %s' % (path, e))

    
    purge_old_logs(path, mylog)
    
    log_path = os.path.join(path, my_pid + '.log')
    try:
        handler = ConcurrentRotatingFileHandler(log_path, maxBytes=4194304,backupCount=7)
    except Exception, e:
        warn('PiCloud cannot open pid-logfile at %s.  Error is %s' % (log_path, e))
    else:
        #hack for SUDO user
        fix_sudo_path(log_path)
        handler.setFormatter(logging.Formatter(logfmt_main, datefmt =datefmt))
        mylog.addHandler(handler)
    

"""Initialize logging"""
def _init_logging():    
    import errno
    from warnings import warn
    
    mylog = logging.getLogger("Cloud")    
    
    #clear handlers if any exist
    handlers = mylog.handlers[:]
    for handler in handlers:
        mylog.removeHandler(handler)
        handler.close()
        
    if saveLog:              
                                  
        from .util.cloghandler.cloghandler import ConcurrentRotatingFileHandler, concurrent_error
        path = os.path.expanduser(cc.baselocation)
        try:
            os.makedirs(path)
        except OSError, e: #allowed to exist already
            if e.errno != errno.EEXIST:                
                warn('PiCloud cannot create directory %s. Error is %s' % (path, e))
        
        path = os.path.join(path, logFilename)
        
        try:
            handler = ConcurrentRotatingFileHandler(path,maxBytes=4194304,backupCount=7)
        except Exception, e: #warn on any exception            
            warn('PiCloud cannot open logfile at %s.  Error is %s' % (path, e))
            handler = NullHandler()
        else:
            fix_sudo_path(path)
            fix_sudo_path(path.replace('.log', '.lock'))
    else:
        #need a null hander
        handler = NullHandler()
    handler.setFormatter(logging.Formatter(logfmt_pid, datefmt=datefmt))
    mylog.addHandler(handler)
    mylog.setLevel(logging.getLevelName(logLevel))    
    
    #start console logging:
    printhandler = logging.StreamHandler()
    printhandler.setLevel(logging.getLevelName(printLogLevel))
    printhandler.setFormatter(
       logging.Formatter(logfmt_main,
       datefmt= datefmt))
    mylog.addHandler(printhandler)
    
    #If pilog exists, use it as a parent
    try:
        pilog_module = __import__('pimployee.log', fromlist=['log'])
        pilogger = pilog_module.pilogger
        mylog.parent = pilogger
    except (ImportError, AttributeError):
        pass
        
    
        
    if not isinstance(handler, NullHandler):
        mylog.debug("Log file (%s) opened" % handler.baseFilename)
        
        # with pid based logging, this is less relevant now
        if concurrent_error:
            mylog.warning('Could not use ConcurrentRotatingFileHandler due to import error: %s.' +
                          'Likely missing pywin32 packages. Falling back to regular logging; you may experience log corruption if you run multiple python interpreters',
                          concurrent_error)        

    if saveLog:
        init_pid_log(os.path.expanduser(cc.baselocation), mylog)

            
    return mylog
                                                              
scicloudLog = _init_logging()


"""verbose mode
Whether the below functions are sent
"""
verbose = cc.logging_configurable('verbose',
                                     default=False, 
                                     comment = "Should scicloud library print informative messages to stdout and stderr",
                                     hidden = True)

def stdout(s, auto_newline=True):
    """Write to stdout if verbose
    If auto_newline, add \n to end (like print)"""
    if not verbose:
        return 
    sys.stdout.write(s)
    if auto_newline:
        sys.stdout.write('\n')
    sys.stdout.flush()
               
def stderr(s, auto_newline=True):
    """Write to stderr if verbose
    If auto_newline, add \n to end (like print)"""
    if not verbose:
        return 
    sys.stderr.write(s)
    if auto_newline:
        sys.stderr.write('\n')
    sys.stderr.flush()



"""
Science VM volume management.
This module allows the user to manage their volumes on PiCloud.
See documentation at http://docs.scivm.com

Be advised that you must have rsync installed for this module to work
"""
from __future__ import absolute_import
"""
Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""
try:
    import json
except:
    # Python 2.5 compatibility
    import simplejson as json
import logging
import os
import platform
import time

import scicloud
from .scicloudlog import stdout as print_stdout, stderr as print_stderr
from .util import common

scicloudLog = logging.getLogger('Cloud.volume')

plat = platform.system()

_urls = {'list': 'volume/list/',
         'create': 'volume/create/',
         'mkdir': 'volume/mkdir/',
         'sync_initiate': 'volume/sync_initiate/',
         'sync_terminate': 'volume/sync_terminate/',
         'delete': 'volume/delete/',
         'check_release': 'volume/check_release/',
         'ls': 'volume/ls/',
         'rm': 'volume/rm/'
         }

_volume_path_delimiter = ':'

_SYNC_READY = 'ready'
_SYNC_NOVACANCY = 'novacancy'
_SYNC_ERROR = 'error'

_RELEASE_DONE = 'done'
_RELEASE_IN_PROGRESS = 'waiting'
_RELEASE_ERROR = 'error'


def _send_vol_request(request_type, data, jsonize_values=True):
    type_url = _urls.get(request_type)
    if type_url is None:
        raise LookupError('Invalid vol request type %s' % request_type)
    return common._send_request(type_url, data, jsonize_values)

"""
volume management
"""
def get_list(name=None, desc=False):
    """Returns a list of dictionaries describing user's volumes.
    If *name* is specified, only shows info for the volume with that name.
    If *desc* is True (default=False), then the description is also displayed.

    Volume information is returned as list of dictionaries.  The keys within
    each returned dictionary are:

    * name:
        name of the volume
    * desc:
        description of the volume (if desc option is True)
    * mnt_path:
        filesystem path where volume contents can be accessed by a job
    * created:
        time when the volume was created
    """
    v_list = _send_vol_request('list', {'name': name, 'desc': desc})
    return [common._fix_time_element(v, 'created') for v in v_list['volumes']]

def create(name, mount_path, desc=None):
    """Creates a new scicloud volume.
        * name:
            name of the new volume (max 64 chars)
        * mount_path:
            If an absolute path is specified, that path is where this volume
            will be mounted when jobs are run specifying access to this volume,
            i.e. mount point where jobs can access the contents of this volume.
            If a relative path is specified, then the mount point is the
            specified path relative to /home/scivm, which is the directory
            where all jobs initially start.
        * desc:
            (optional) description of the volume (max 1024 chars)
    """
    if len(name) < 2:
        raise scicloud.CloudException('Volume name must be at least 2 characters.')
    _send_vol_request('create',
                      {'name': name, 'mnt_path': mount_path,
                       'desc': desc or ''})
    scicloudLog.debug('created volume %s', name)

def mkdir(volume_path, parents=False):
    """Creates directory(ies) at volume_path, if they don't already exist.

    * volume_path:
        A scicloud volume path spec or a list of specs, that indicates the
        directory(ies) to create.
    * parents:
        If True, does not error if the directory already exists, and makes any
        necessary parent directories.
    """
    vol_name, vol_paths = common.parse_remote_paths(volume_path)
    res = _send_vol_request('mkdir',
                            {'name': vol_name, 'paths': vol_paths,
                             'parents': parents})

    if res.get('modified'):
        _wait_for_release(vol_name)
    msg = 'created %s in volume %s' % (', '.join(vol_paths), vol_name)
    scicloudLog.debug(msg)
    print_stdout(msg)

def sync(source, dest, delete=False):
    """Syncs data between a scicloud volumes and the local filesystem.

    Either *source* or *dest* should specify a scicloud volume path, but not both.
    A scicloud volume path is of the format:

        volume_name:[path-within-volume]

    where path-within-volume cannot be an absolute path (There is no concept of
    the root of the filesystem in a volume: All path specifications are relative
    to the top level of the volume).  Note that the colon is what indicates this
    is a volume path specification.  Local paths should point to a local
    directory or file.  If the local path is a directory, whether the
    directory itself or the contents of the directory are synced depends on the
    presence of a trailing slash.  A trailing slash indicates that the contents
    should be synced, while its absence would lead to the directory itself being
    synced to the volume.  *source* can be a list of paths, all of which should
    either be local paths, or volume paths in the same scicloud volume.

    Example::

        sync('~/dataset1', 'myvolume1:')

    will ensure that a directory named 'dataset1' will exist at the top level
    of the scicloud volume 'myvolume1', that contains all the contents of
    'dataset1'.  On the other hand,

        sync('~/dataset1/', 'myvolume1:')

    will copy all the contents of 'dataset1' to the top level of 'myvolume1'.
    This behavior mirrors the file-copying tool 'rsync'.
    
    If *delete* is True, files that exist in *dest* but not in *source* will be
    deleted.  By default, such files will not be removed.
    """
    conn = scicloud._getscicloudnetconnection()
    retry_attempts = conn.retry_attempts
    dest_is_local = common.is_local_path(dest)
    l_paths, r_paths = (dest, source) if dest_is_local else (source, dest)
    local_paths = common.parse_local_paths(l_paths)
    vol_name, vol_paths = common.parse_remote_paths(r_paths)
    for vol_path in vol_paths:
        if os.path.isabs(vol_path):
            raise scicloud.CloudException('Volume path cannot be absolute')
    
    # acquire syncslot and syncserver info to complete the real remote paths
    success = release = False
    exit_code = -1
    syncserver, syncslot = _acquire_syncslot(vol_name)

    try:
        scicloudLog.debug('Acquired syncslot %s on server %s', syncslot, syncserver)
        r_base = '%s@%s:volume/' % (syncslot, syncserver)
        r_paths = ' '.join(['%s%s' % (r_base, v_path) for v_path in vol_paths])
        l_paths = ' '.join(local_paths)
        sync_args = (r_paths, l_paths) if dest_is_local else (l_paths, r_paths)

        for attempt in xrange(retry_attempts):
            exit_code, stdout, stderr = common.rsync_session(*sync_args,
                                                             delete=delete)
            if not exit_code:
                break
            scicloudLog.error('sync attempt failed:\n%s', stderr)
            print_stdout(str(stderr))
            print_stdout('Retrying volume sync...')
        else:
            raise Exception('sync failed multiple attempts... '
                            'Please contact PiCloud support')
    except KeyboardInterrupt:
        scicloudLog.error('Sync interrupted by keyboard')
        print 'Sync interrupted by keyboard'
    except Exception as e:
        scicloudLog.error('Sync errored with:\n%s', e)
        print e
    finally:
        print_stdout('Cleanup...')
        success = not exit_code
        release = success and not dest_is_local
        _send_vol_request('sync_terminate', {'name': vol_name,
                                             'syncslot': syncslot,
                                             'syncserver': syncserver,
                                             'release': release})

    if release:
        print_stdout('Ensuring redundancy...')
        _wait_for_release(vol_name)
    if success:
        print_stdout('Sync successfully completed.')
    else:
        raise scicloud.CloudException('Volume sync failed with error code %s. '
                                   'See scicloud.log' % exit_code)

def delete(name):
    """Deletes the scivm volume identified by *name*."""
    _send_vol_request('delete', {'name': name})
    scicloudLog.debug('deleted volume %s', name)

def ls(volume_path, extended_info=False):
    """Lists the contents at *volume_path*.
    
    * volume_path:
        A scicloud volume path spec or a list of specs, whose contents are to be
        returned.
    * extended_info:
        If True, in addition to the names of files and directories comprising
        the contents of the volume_path, the size (in bytes) and the modified
        times are returned. (Default is False)
    
    Returns a list of tuples, one for each volume path specified.  The first
    element of the tuple is the volume path spec, and the second element of the
    tuple is a list of dictionaries for each file or directory present in the
    volume path.
    """
    vol_name, vol_paths = common.parse_remote_paths(volume_path)
    res = _send_vol_request('ls',
                            {'name': vol_name, 'paths': vol_paths,
                             'extended_info': extended_info})

    fixed_listings = []
    for v_path, listings in res.get('listings'):
        v_path = '%s:%s' % (vol_name, v_path)
        if extended_info:
            listings = [common._fix_time_element(v, 'modified') for v in listings]
        fixed_listings.append((v_path, listings))

    return fixed_listings

def rm(volume_path, recursive=False):
    """Removes contents at *volume_path*.
    
    * volume_path:
        A scicloud volume path spec or a list of specs, whose contents are to be
        removed.
    * recursive:
        If True, will remove the contents at *volume_path* recursively, if it
        is a directory.  If *recursive* is False, and *volume_path* points to
        a non-empty directory, it is an error. (Default is False)
    """
    vol_name, vol_paths = common.parse_remote_paths(volume_path)

    res = _send_vol_request('rm',
                            {'name': vol_name,
                             'paths': vol_paths,
                             'recursive': recursive})
    if res.get('modified'):
        _wait_for_release(vol_name)
    scicloudLog.debug('removed %s from volume %s', ', '.join(vol_paths), vol_name)

def _acquire_syncslot(volume_name):
    """Requests syncslot from PiCloud.  Current behavior is to try 12 times,
    waiting 5 seconds between failed attempts."""
    num_retries = 12
    wait_time = 5   # seconds
    print_stdout('Connecting with PiCloud to initiate sync', False)
    while num_retries:
        print_stdout('.', False)
        res = _send_vol_request('sync_initiate', {'name': volume_name})
        status = res.get('status')
        if status == _SYNC_NOVACANCY:
            num_retries -= 1
            time.sleep(wait_time)
            continue
        if status not in [_SYNC_READY, _SYNC_ERROR]:
            status = _SYNC_ERROR
        break
    print_stdout('')
    
    if status == _SYNC_NOVACANCY:
        scicloudLog.error('No available syncslot')
        raise scicloud.CloudException('Volume sync is unavailable at the moment.  '
                                   'Please try again in a few minutes.  '
                                   'We Apologize for the inconvenience.')
    if status == _SYNC_ERROR:
        scicloudLog.error('Error acquiring syncslot')
        raise scicloud.CloudException('Could not complete volume sync.  '
                                   'Please contact PiCloud support.')

    return res.get('syncserver'), res.get('syncslot')

def _wait_for_release(volume_name, wait_interval=3):
    """Polls volume's status until it's no longer waiting release."""
    while True:
        res = _send_vol_request('check_release', {'name': volume_name})
        status = res['status']
        if status == _RELEASE_ERROR:
            raise scicloud.CloudException('Sync failed on volume %s' % volume_name)
        if status == _RELEASE_DONE:
            break
        time.sleep(3)



"""
Defines all Cloud Adapters - objects that control lower-level behavior of the scicloud
There is a one to one instance mapping between a scicloud and its adapter

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from __future__ import with_statement

import sys
import time
import types
import threading
from functools import partial
from itertools import izip, imap, count

from ..scicloud import CloudException
from .. import serialization
from .. import scicloudconfig as cc
from ..scicloudlog import scicloudLog
from ..util import OrderedDict


class Adapter(object):
    """
    Abstract class to deal with lower-level scicloud operations
    """
          
    _isopen = False
    
    @property
    def opened(self): 
        """Returns whether the adapter is open"""
        return self._isopen
    
    def open(self):
        """called when this adapter is to be used"""
        if self.opened:
            raise CloudException("%s: Cannot open already-opened Adapter", str(self))
        if not self._scicloud.opened:
            self._scicloud.open()
    
    def close(self):
        """called when this adapter is no longer needed"""
        if not self.opened:
            raise CloudException("%s: Cannot close a closed adapter", str(self))
        self._isopen = False
        
    @property
    def scicloud(self):
        return self._scicloud
    
    def needs_restart(self, **kwargs):
        """Called to determine if the scicloud must be restarted due to different adapter parameters"""
        return False
        
    def call(self, params, func, *args):
        raise NotImplementedError
    
    def jobs_join(self, jids, timeout = None):
        """
        Allows connection to manage joining
        If connection manages joining, it should return a list of statuses  
        describing the finished job
        Else, return False
        """
        return False
    
    def jobs_map(self, params, func, argList):
        """Treat each element of argList as func(*element)"""
        raise NotImplementedError
    
    def jobs_result(self, jids):
        raise NotImplementedError

    def jobs_kill(self, jids):
        raise NotImplementedError
    
    def jobs_delete(self, jids):
        raise NotImplementedError    
    
    def jobs_info(self, jids, info_requested):
        raise NotImplementedError
    
    def is_simulated(self):        
        raise NotImplementedError
    
    def connection_info(self):
        return self.connection.connection_info()
    
    def dep_snapshot(self):
        """This function, called downstream from the network connection object, 
        deals with calling the dependency checking functions
        Note that it can be called in between successive maps"""
        pass #do nothing by default

        

"""
TheSerializingAdapter is responsible for serializing data passed into args
It then passes data onto its CloudConnection handler
It can (optionally) log serializations
"""

class SerializingAdapter(Adapter):
    
    #config for pickle debugging        
    c = """Should serialization routines track debugging information?
    If this is on, picklingexceptions will provide a 'pickle trace' that identifies what member of an object cannot be pickled"""
    serializeDebugging =  \
        cc.logging_configurable('serialize_debugging',
                                          default=True,
                                          comment=c, hidden=True)
    c = """Should all object serialization meta-data be logged to the serialize_logging_path?""" 
    serializeLogging = \
        cc.logging_configurable('serialize_logging',
                                          default=False,
                                          comment=c)      
    if serializeLogging and not serializeDebugging:
        serializeDebugging = True
        scicloudLog.warning("serialize_logging implies serialize_debugging. Setting serialize_debugging to true.")
    
    
    c = """Maximum amount of data (in bytes) that can be sent to the PiCloud server per function or function argument. 
    Value may be raised up to 16 MB"""
    max_transmit_data = \
        cc.transport_configurable('max_transmit_data',
                                  default=1000000,
                                  comment=c)
    max_transmit_data = min(max_transmit_data,16000000)
        
    
    
    #live:
    _report = None  
    _isSlave = False #True if this adapter is a slave (it's connection is in a master process)
                  
        
    def __init__(self, connection, isSlave = False):
        self.connection = connection 
        connection._adapter = self
        self.opening = False
        self.openLock = threading.RLock()            
        self._isSlave = isSlave
        
    def init_report(self, subdir = ""):
        try:
            self._report = serialization.SerializationReport(subdir)
        except IOError, e:
            scicloudLog.warn('Cannot construct serialization report directory. Error is %s' % str(e))
            self._report = None
            self.serializeLogging = False #disable logging
        
    @property
    def report(self):
        return self._report
    
    @property
    def isSlave(self):
        return self._isSlave

    def _configure_logging(self):
        #Seperated from open to allow for dynamic changes at runtime
        
        #set up scicloud serializer object
        self.min_size_to_save =  0 if self.serializeLogging else (self.max_transmit_data / 3)   
        #reporting:                
        if not self._report:
            if self.serializeLogging or self.connection.force_adapter_report():
                if self._isSlave:
                    self.init_report()
                    self._report.logPath = self.connection.get_report_dir() 
                else:
                    self.init_report(self.connection.report_name())            
            else:
                self._report = None  

    def getserializer(self, serialization_level = 0):
        """Return the correct serializer based on user settings and serialization_level"""
        if serialization_level >= 2:
            return serialization.Serializer
        elif self.serializeDebugging and serialization_level == 0:
            return serialization.DebugSerializer
        else:
            return serialization.CloudSerializer        
        
    
    def open(self):      
        with self.openLock:
            if self._isopen or self.opening:
                return

            self.opening = True
            Adapter.open(self)  
                                                
            try:                 
                #always try to open               
                if self._isSlave:  
                    #must open connection first to get reportDir                  
                    self.connection.open()
                    self._configure_logging()
                else:
                    self._configure_logging()
                    self.connection.open()
            finally:
                self.opening = False        
                        
            self._isopen = True
        
        
                 
    def close(self):
        Adapter.close(self)
        if hasattr(self.connection, 'opened') and self.connection.opened:
            self.connection.close()        

    def check_size(self, serializerInst, logfilename, is_result=False, size_limit = None):
        """Check if transmission size exceeds limit"""
        if size_limit is None:
            size_limit = self.max_transmit_data
        totalsize = len(serializerInst.serializedObject)
        if totalsize > self.max_transmit_data:
            exmessage = 'Excessive data (%i bytes) transmitted.  See help at %s' % (totalsize, 'http://docs.scivm.com')
            if hasattr(serializerInst,'str_debug_report'):
                exmessage += ' Snapshot of what you attempted to %s:\n' % ('return' if is_result else 'send')
                serializerInst.set_report_minsize(totalsize/3)
                exmessage += serializerInst.str_debug_report(hideHeader=True)
            else:
                exmessage += '\n'
            exmessage += 'Cloud has produced this error as %s too large of an object.\n' %('your job is returning' if is_result else 'you are sending')
            if hasattr(self,'str_debug_report'):
                if is_result:
                    exmessage += 'The above snapshot describes the data that would be returned by this job.\n'
                else:
                    exmessage += 'The above snapshot describes the data that must be transmitted to execute your PiCloud call.\n'
            else:
                exmessage += '\n To see data snapshots, use _fast_serialization = 0 (default)'
                if not cc.transport_configurable('running_on_scicloud',default=False):
                    exmessage += ' and enable serialize_debugging in scicloudconf.py\n'
                else:
                    exmessage += '\n'
            if is_result:
                exmessage += 'You cannot return more than %s MB from a job' % (size_limit / 1000000)
            elif cc.transport_configurable('running_on_scicloud',default=False):
                exmessage += 'You cannot send more than %s MB per job argument via scicloud.call/map' % (size_limit / 1000000)
            else:
                exmessage += 'If you decide that you actually need to send this much data, increase max_transmit_data in scicloudconf.py (max 16 MB)\n'
             
            if logfilename:
                exmessage+= 'See entire serialization snapshot at ' + logfilename
            elif not self.scicloud.running_on_scicloud():
                exmessage+='Please enable serialize_logging in scicloudconf.py to see more detail'
            raise CloudException(exmessage)
    
    def _scicloud_serialize_helper(self, func, arg_serialization_level, args, argnames=[], logprefix="", 
                                coerce_cnt=None, os_env_vars=[]):
        """Returns a serialization stream which produces a serialized function, then its arguments in order
        Also returns name of logfile and any counter associated with it
        """
        baselogname =  logprefix

        if isinstance(func,partial):
            f = func.func
        else:
            f = func
        
        if f:
            if hasattr(f,'__module__'):            
                baselogname+=(f.__module__ if f.__module__ else '__main__') +'.'
            if hasattr(f,'__name__'):
                baselogname+=f.__name__        
        
        acname = None                
      
        sargs = []
        
        if coerce_cnt != None:
            cnt = coerce_cnt
        else:
            cnt = self._report.update_counter(baselogname) if self._report else 0

        def generate_stream():
            """Helper generator that produces the stream"""
            #serialize function:
            if f:
                sfunc = self.getserializer(0)(func)
                sfunc.set_os_env_vars(os_env_vars)
                try:
                    sfunc.run_serialization(self.min_size_to_save)
                finally:        
                    if self.serializeLogging:
                        logname = baselogname + '.%sfunc'
                        acname = self._report.save_report(sfunc, logname, cnt)
                    else:
                        acname = ""  
                
                self.check_size(sfunc,acname) 
                yield sfunc
            
            #arguments
            if args:
                argSerializer = self.getserializer(arg_serialization_level)
                for obj, name in izip(args, argnames):
                    #TODO: Policy change here
                    serializerI = argSerializer(obj)
                
                    try:
                        serializerI.run_serialization(self.min_size_to_save)
                    finally:
                        if self.serializeLogging:                        
                            logname = baselogname + '.%s' + name
                            acname = self._report.save_report(serializerI, logname, cnt)  
                        else:
                            acname = ""
                
                    self.check_size(serializerI,acname) 
                
                    yield serializerI                                        
        return generate_stream(), baselogname, cnt

        
    
    def scicloud_serialize(self, func, arg_serialization_level, args, argnames=[], logprefix="", 
                        coerce_cnt=None, os_env_vars=[]):
        """Return serialized_func, list of serialized_args
        Will save func and args to files.
        """
        
        serialize_stream, logname, cnt = self._scicloud_serialize_helper(func, arg_serialization_level, 
                                                                      args, argnames, logprefix, 
                                                                      coerce_cnt, os_env_vars)
        if func:
            sfunc = serialize_stream.next().serializedObject
        else:
            sfunc = None        
        if args:
            sargs = imap(lambda sarg: sarg.serializedObject, serialize_stream) #stream
        else:
            sargs = None
                
        return sfunc, sargs, logname, cnt
    
    
    def map_reduce_job(self, mapper_func, reducer_func, bigdata_file):
        
        # serialize the the args above into params
        smapper, _, _, _ = self.scicloud_serialize(mapper_func, 0, [])
        sreducer, _, _, _ = self.scicloud_serialize(reducer_func, 0, [])
        
        params = {}
        params['mapper_func'] = smapper
        params['reducer_func'] = sreducer
        params['bigdata_file'] = bigdata_file
        
        return self.connection.add_map_reduce_job(params)

    
    def job_call(self, params, func, args, kwargs):
        os_env_vars = params.pop('os_env_vars', None)        
        sfunc, sargs, logprefix, logcnt =  self.scicloud_serialize(func, params['fast_serialization'], 
                                                                [args, kwargs], ['args', 'kwargs'], 'call.',
                                                                os_env_vars=os_env_vars) 
        
        params['func'] = sfunc
        params['args'] = sargs.next()
        params['kwargs'] = sargs.next()

        return self.connection.job_add(params=params, logdata = (logprefix, self._report.pid if self._report else 0, logcnt))
    
    def jobs_join(self, jids, timeout = None):
        return self.connection.jobs_join(jids, timeout)
    
    def jobs_map(self, params, func, mapargs, mapkwargs = None):
        """Treat each element of mapargs as func(*element)
        Mapkwargs implementation is a bit inefficient in that it can call two 
        module_add/check. however, this will be a very rare case"""
        
        mapargnames = imap(lambda x: 'jobarg_' + str(x),count(1)) #infinite stream generates jobarg_ for logging
        mapkwargnames = imap(lambda x: 'jobkwarg_' + str(x),count(1)) #infinite stream generates jobarg_ for logging                    
        
        os_env_vars = params.pop('os_env_vars', None)
        sfunc, sargs, logname, logcnt =  self.scicloud_serialize(func, params['fast_serialization'], 
                                                              mapargs, mapargnames, 'map.', 
                                                              os_env_vars=os_env_vars)
        #handle kwargs if present
        if mapkwargs:
            _, skwargs, _, _ = self.scicloud_serialize(None, params['fast_serialization'], mapkwargs, 
                                                    mapkwargnames, logname, coerce_cnt = logcnt)
        else:
            skwargs = None        

        params['func'] = sfunc
        
        if self.isSlave: #fixme: Find a way to pass stream through to master
            sargs = list(sargs)
            
        return self.connection.jobs_map(params=params, mapargs = sargs, mapkwargs = skwargs,
                                       logdata = (logname, self._report.pid if self._report else 1, logcnt))
        
    def jobs_result(self, jids, by_jid):
        """Returns serialized result; higher level deals with it"""
        return self.connection.jobs_result(jids=jids, by_jid=by_jid)
    
    def jobs_kill(self, jids):
        return self.connection.jobs_kill(jids=jids)
    
    def jobs_delete(self, jids):
        return self.connection.jobs_delete(jids=jids)        
    
    def jobs_info(self, jids, info_requested):
        return self.connection.jobs_info(jids=jids, info_requested=info_requested)
   
    def is_simulated(self):        
        return self.connection.is_simulated()
    
    def needs_restart(self, **kwargs):        
        return self.connection.needs_restart(**kwargs)

class DependencyAdapter(SerializingAdapter):
    """
    The DependencyAdapter is a SerializingAdapter capable of sending modules
    to PiCloud
    """
    

    # manages file dependencies:
    dependencyManager = None


    c = """Should a more aggressive, less optimal, module dependency analysis be used?  This is always used on stdin"""
    aggressiveModuleSearch = \
        cc.transport_configurable('aggressive_module_search',default=False,comment=c, hidden=True)

    c = """Set to False to disable automatic dependency transfer. This will not affect dependencies already sent to PiCloud."""
    automaticDependencyTransfer = \
        cc.transport_configurable('automatic_dependency_transfer',default=True,comment=c, hidden=False)


    def _check_forced_mods(self):
        """find things that may be imported via import a.b at commandline
        Set __main__.___pyc_forcedImports__ to that list
        """
        marked = set()        
        main = (sys.modules['__main__'])  
        marked.add(main)        
        def recurse(testmod, name): #name is tuple version of testmod
            for key, val in testmod.__dict__.items():                
                if not name: #inspecting main module -- use val name
                    if isinstance(val,types.ModuleType):
                        item = val.__name__
                        if item in ['scicloud','__builtin__']:
                            continue
                        tst = tuple(item.split('.'))
                        
                    else:
                        continue
                else: #child module cannot be renamed inside a parent
                    item = key
                    tst = name + (item,) 
                if item in marked: 
                    continue                
                tstp = '.'.join(tst)                
                themod = sys.modules.get(tstp)
                if not themod:
                    continue
                
                self._create_dependency_manager()
                if self.dependencyManager:
                    self.dependencyManager.inject_module(themod)  
                marked.add(item)
                if len(tst) > 1: #this is a forced import:
                    if not hasattr(main,'___pyc_forcedImports__'):
                        main.___pyc_forcedImports__ = set()
                    main.___pyc_forcedImports__.add(themod)
                recurse(themod,tst)                                    
        with self.modDepLock:
            recurse(main,tuple())                

    
    def _serialize_dep_handler(self, sarg):
        """Used by scicloud_serialize imap on serialized args"""
        
        self._create_dependency_manager()
        if self.dependencyManager:
            with self.modDepLock:
                for moddep in sarg.get_module_dependencies():
                    self.dependencyManager.inject_module(moddep)
        return sarg.serializedObject
        
    
    def scicloud_serialize(self, func, arg_serialization_level, args, argnames=[], logprefix="", 
                        coerce_cnt=None, os_env_vars=[]):
        """
        Similar to the normal scicloud_serialize, except tracks dependencies
        """
        
        try:
            self._create_dependency_manager()
                           
            main = (sys.modules['__main__'])
            main_file = getattr(main, '__file__', None)
            # TODO: ipython support?
            force_aggressive = not main_file or main_file == '<stdin>'
            
            if serialization.scicloudpickle.useForcedImports and \
                (self.aggressiveModuleSearch or force_aggressive):
                self._check_forced_mods()      
    
            if not self.automaticDependencyTransfer:
                return SerializingAdapter.scicloud_serialize(self, func, arg_serialization_level, 
                                                          args, argnames, logprefix, coerce_cnt,
                                                          os_env_vars)   
            
            serialize_stream, logname, cnt = self._scicloud_serialize_helper(func, arg_serialization_level, 
                                                                          args, argnames, logprefix, 
                                                                          coerce_cnt, os_env_vars)
                    
            if func:
                sfunc = serialize_stream.next()
                outfunc = self._serialize_dep_handler(sfunc)
            else:
                func = None
                outfunc = None
                        
            if args:
                outargs = imap(lambda sarg: self._serialize_dep_handler(sarg), serialize_stream)
            else:
                outargs = None
    
            return outfunc, outargs, logname, cnt
        
        except Exception, e: # report exceptions
            
            
            try:
                from .network import HttpConnection, scicloudLog as conLog
                import platform
                import traceback
                
                if isinstance(self.connection, HttpConnection):
                    self.connection.send_request('report/python_error/', 
                                                 {'exception' : str(e), 
                                                  'stacktrace' : ''.join(traceback.format_stack()),
                                                  'traceback' : ''.join(traceback.format_tb(sys.exc_info()[2])),                                              
                                                  'hostname': platform.node(),
                                                  'language_version': platform.python_version(),
                                                  'language_implementation': platform.python_implementation(),
                                                  'platform': platform.platform(),
                                                  'architecture': platform.machine(),
                                                  'processor': platform.processor(),
                                                  'pyexe_build' : platform.architecture()[0]
                                                      }, logfunc=conLog.debug )                
                
            except Exception:
                pass # absolutely do not let an exception corrupt user seeing the true error                
            raise # propagate old exception 

    def dep_snapshot(self):
        """This function, called downstream from the network connection object, 
        deals with checking for new modules and sending them over the network
        
        It does not return anything"""
        
        self._create_dependency_manager()
        if not self.dependencyManager:
            return
        
        with self.modDepLock:
            #lock is held entire time as small chance webserver will raise CloudException            
            deps, new_snapshot = self.dependencyManager.get_updated_snapshot()        
            
            if deps:
            #if False:
                scicloudLog.debug('New Dependencies %s' % str(deps))
                #modules are a tuple of (modname, timestamp, archvie)
                modules = self.connection.modules_check(deps)
                if modules:
                    from ..transport.codedependency import FilePackager
                    f = FilePackager(map(lambda module: (module[0], module[2]), modules), self.dependencyManager)
                    tarball = f.get_tarball()
                    scicloudLog.info('FileTransfer: Transferring %s' % modules)
                    self.connection.modules_add(modules, tarball)
            
            #if no exception raised, commit snapshot
            self.dependencyManager.commit_snapshot(new_snapshot)
    
    def _create_dependency_manager(self, whitelist=None):
        
        if not self.automaticDependencyTransfer:        
            return
        elif not self.dependencyManager:
            from ..transport import DependencyManager
            ignoreList = self.connection.packages_list()
            self.dependencyManager = DependencyManager(excludes=ignoreList, whitelist=whitelist)
    
    def open(self):                
        with self.openLock:
            if self.opening or self.opened:
                return
            SerializingAdapter.open(self)
            if self._scicloud.running_on_scicloud(): 
                # within scicloud, we have all needed dependencies so no need to transfer any
                self.automaticDependencyTransfer = False            
            self.modDepLock = threading.Lock()            
            self._isopen = True



"""
A module provided that emulates python multiprocessing's Process Pool interface.
This can be used to trivially replace multiprocessing Pool code.

Note that scicloud.setkey must still be used before using this interface.

Not supported (mostly due to not making sense in the PiCloud context).

* Pool initializer
* Passing keyword arguments to function being applied
* callback for map/map_async
* (i)map chunksize -- argument only controls scicloud.iresult chunksize
* iresult_undordered -- this is just an ordered iresult
* multiprocessing.TimeoutError -- scicloud.CloudTimeoutError is used instead
* join, close and terminate operations

.. warning::
    
    It is highly recommended that you do not use this module and instead use the scicloud module directly.
    Much functionality that the scicloud interface offers is missing here.
"""
"""
Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from scicloud import CloudTimeoutError


from . import _getscicloud
import multiprocessing

class AsyncResult(object):
    """Result object that emulates multiprocessing.pool.AsyncResult"""
    
    _jid = None #internal - jid (or jid list) associated with this result
    
    def __init__(self, jid):
        self._jid = jid
    
    def get(self, timeout=None):
        """
        Return result when it arrives.
        If timeout is not None and none arrives, 
        raise multiprocessing.TimeoutError in *timeout* seconds
        """
        return _getscicloud().result(self._jid)
        
    def wait(self, timeout=None):
        """
        Wait until result is available or *timeout* seconds pass
        """
        try:
            _getscicloud().join(self._jid)
        except CloudTimeoutError:
            pass
        
    def ready(self):
        """Returns true if the job finished (done or errored)"""
        c = _getscicloud()
        status = c.status(self._jid)
        if not hasattr(status, '__iter__'):
            return status in c.finished_statuses
        else:
            for s in status:
                if s not in c.finished_statuses:
                    return False
            return True
    
    def successful(self):
        """Returns true if job finished successfully.
        Asserts that job has finished"""
        assert(self.ready())
        status = _getscicloud().status(self._jid)
        if not hasattr(status, '__iter__'):
            return status == 'done'
        else:
            for s in status:
                if s != 'done':
                    return False
            return True

    

def apply(func, args=()):
    """
    Equivalent to Multiprocessing apply.
    keyword arguments are not supported        
    """
    
    c = _getscicloud()
    jid = c.call(func, *args)
    return c.result(jid)

def apply_async(func, args=(), callback=None):
    """
    Equivalent to Multiprocessing apply_async
    keyword arguments are not supported
    
    callback is a list of functions that should be run on the callee's computer once this job finishes successfully.
        Each callback will be invoked with one argument - the jid of the complete job     
    """

    c = _getscicloud()
    jid = c.call(func, _callback = callback, *args)
    return AsyncResult(jid)
    

def map(func, iterable, chunksize=None):
    """
    Equivalent to Multiprocessing map
    chunksize is not used here
    """
    
    c = _getscicloud()
    jids = c.map(func, iterable)
    return c.result(jids)

def map_async(func, iterable, chunksize=None):
    """
    Equivalent to Multiprocessing map_async
    chunksize is not used here
    """    
    c = _getscicloud()
    jids = c.map(func, iterable)
    return AsyncResult(jids)
    
    
def imap(func, iterable, chunksize = None):
    """
    Equivalent to Multiprocessing imap
    chunksize is used only to control the scicloud.iresult stage
    """
    
    c = _getscicloud()
    jids = c.map(func, iterable)
    return c.iresult(jids,chunksize)

def imap_unordered(func, iterable, chunksize = None):
    """
    Same as imap
    """
    
    return imap(func, iterable, chunksize)

    


    
    


"""
PiCloud configuration settings
Two settings are stored in this file - the baseLocation for scicloud AND the config file location
All other variables are stored in a config file managed by this

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""


from __future__ import with_statement
import os
import distutils
import distutils.dir_util

#Location for all scicloud configuration
if os.name == "nt":
    baselocation = os.environ.get('APPDATA')
    if baselocation:
        baselocation = os.path.join(baselocation, 'scivm')
    else:
        baselocation =  os.path.join('~', '.scivm')
else:
    baselocation = os.path.join('~', '.scivm')


#name of configuration file:
configname = 'scicloudconf.py'

_needsWrite = False

genHidden = False #if true config will generate hidden variables

from .util import configmanager

def get_config_value(section, varname, default, comment=None, hidden=False):
    
    """
    Return value associated with a varname found in section
    If varname is not found, return defaultval and add varname
    If varname found, cast value to type of default
    """       
    typ = type(default)
    try:
        entry = config.get(section, varname, comment)
        return typ(entry)
        
    except (configmanager.NoOptionError), e:        
        if hidden and not genHidden:
            config.hiddenset(section,varname,default,comment)            
            return default
        config.set(section,varname,default,comment)        
        return default
    except ValueError, e:
        import logging
        log = logging.getLogger('Cloud') #might not work if logging not yet initialized
        log.warning('Option %s.%s in scicloudconf.py must have type %s. Reverting to default' % (section, varname, typ.__name__))
        return default    

"""Sections defined below"""    
def account_configurable(varname,default,comment=None, hidden=False):
    return get_config_value('Account',varname,default,comment,hidden)
    
def logging_configurable(varname,default,comment=None, hidden=False):
    return get_config_value('Logging',varname,default,comment,hidden)

def mp_configurable(varname,default,comment=None, hidden=False):
    return get_config_value('Multiprocessing',varname,default,comment,hidden)

def simulation_configurable(varname,default,comment=None, hidden=False):
    return get_config_value('Simulation',varname,default,comment,hidden)

def transport_configurable(varname, default, comment=None, hidden=False):
    return get_config_value('Transport',varname,default,comment,hidden)
    


def flush_config():
    """
    Write settings to file
    """
    conf_path = os.path.join(fullconfigpath,configname)
    try:
        with open(conf_path, 'w') as configfile:    
            config.write(configfile)
    except IOError, e:
        import logging
        log = logging.getLogger('Cloud') #might not work if logging not yet initialized
        log.exception('Could not write %s.', conf_path)

    

""" Setup
"""
import os
import distutils.errors

#test

config = configmanager.ConfigManager()
fullconfigpath = os.path.expanduser(baselocation)
try:
    newdir = distutils.dir_util.mkpath(fullconfigpath) #ensure path exists
except distutils.errors.DistutilsFileError:
    #this will fail
    config.read(os.path.join(fullconfigpath,configname))

else:
    if newdir:
        os.chmod(newdir[-1], 0700)
        
    if not config.read(os.path.join(fullconfigpath,configname)):
        _needsWrite = True
    #scicloudLog.debug("Cloud Configuration imported")



#!/usr/bin/env python
####
# Heavily modified version of urllib2_file. Original docs follow:
# Version 1.0 (PiCloud):
#  - Support gzip compression each way, use python2.5 http interface, etc.
#  - Add support for persistent HTTP Connections
# Version: 0.2.0
#  - UTF-8 filenames are now allowed (Eli Golovinsky)<br/>
#  - File object is no more mandatory, Object only needs to have seek() read() attributes (Eli Golovinsky)<br/>
#
# Version: 0.1.0
#  - upload is now done with chunks (Adam Ambrose)
#
# TODO: This code is incredibly sloppy. Rewrite one day (Boto connection is better starting point) 

# Version: older
# THANKS TO:
# bug fix: kosh @T aesaeion.com
# HTTPS support : Ryan Grow <ryangrow @T yahoo.com>
# Copyright (C) 2004,2005,2006 Fabien SEISEN
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# you can contact me at: <fabien@seisen.org>
# http://fabien.seisen.org/python/
#
# Also modified by Adam Ambrose (aambrose @T pacbell.net) to write data in
# chunks (hardcoded to CHUNK_SIZE for now), so the entire contents of the file
# don't need to be kept in memory.
#

"""
enable to upload files using multipart/form-data

idea from:
upload files in python:
 http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/146306

timeoutsocket.py: overriding Python socket API:
 http://www.timo-tasi.org/python/timeoutsocket.py
 http://mail.python.org/pipermail/python-announce-list/2001-December/001095.html

import urllib2_files
import urllib2
u = urllib2.urlopen('http://site.com/path' [, data])

data can be a mapping object or a sequence of two-elements tuples
(like in original urllib2.urlopen())
varname still need to be a string and
value can be string of a file object
eg:
  ((varname, value),
   (varname2, value),
  )
  or
  { name:  value,
    name2: value2
  }

"""

import os
import socket
import sys
import stat
import thread
import mimetypes
import errno
#BUG in python 2.6 mimetypes: We must call init() here to ensure thread safety
if not mimetypes.inited:
    mimetypes.init()

import random
import httplib
import urllib
import urllib2

try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO

import logging
scicloudLog = logging.getLogger('Cloud.HTTPConnection')


CHUNK_SIZE = 65536


#PiCloud injection for gzip control
from .. import scicloudconfig as cc

use_gzip = cc.transport_configurable('use_gzip',
                                     default=True,hidden=False,
                                     comment='Request gziped HTTP responses')
client_gzip = False #gzip http requests? (genreally don't use this)

http_close_connection = cc.transport_configurable('http_close',
                                     default=False,hidden=False,
                                     comment='Should every HTTP connection be closed after receiving a response?')
c = """Specify an http/https proxy server that should be used
e.g. proxy.example.com:3128 for anonymous  or username:password@proxy.example.com:3128 for authenticated
"""
proxy_server = cc.transport_configurable('proxy_server',default='',comment = c) 

#new:
from .gzip_stream import GzipFile
#from gzip import GzipFile


def choose_boundary():
    """Choose boundary with more randomness
    Pray that it doesn't conflict :)"""
    import email.generator as generator
    _fmt = generator._fmt
    boundary = ('=' * 8)
    for i in range(8):          
        token = random.randrange(sys.maxint)
        boundary += _fmt % token
    boundary += '=='
    return boundary

def get_content_type(filename):
    return mimetypes.guess_type(filename)[0] or 'application/octet-stream'

def send_data(v_vars, v_files, boundary, gzip = True, sock = None):
    """Either returns data or sends in depending on arguments.
    
    Return value is a tuple of string data, content-length
    
    Content-length always returns; it represents the content-length of the http request
    
    If sock is None:
        Typically string data is returned; this can be written straight to the socket
        
        However, as an optimization, if any v_files correspond to a file on disk and gzip is false,
            The data returned will be none. This must then call send_data again with sock nonnull
    
    If sock is a connection, content-length will be returned AND data will be streamed over socket (but not returned) 
        direct socket writing will not work if gzip is True as content-length of http request will be wrong
    """
    
    all_in_memory = True  #if we can return this
    cl = 0 #holds content_length
    
    if gzip: 
        if sock:
            raise TypeError('gzip and sock cannot both be true')  
        gzip_file = StringIO()
        buffer = GzipFile(filename=None,fileobj=gzip_file,mode='w')
    else:
        buffer = StringIO()

    
    for (k, v) in v_vars:
        buffer.write('--%s\r\n' % boundary)
        buffer.write('Content-Disposition: form-data; name="%s"\r\n\r\n%s\r\n' % (k,v))
    
    for (k, v) in v_files:
        fd = v
        cur_pos = None
        #print 'transmit %s %s' % (k,v)
        contents = None
        try:  #check if String IO
            contents = fd.getvalue()
            file_size = len(contents)
            name = k
        except AttributeError, e:
            #This might be a file on disk
            try:
                
                assert not gzip #if gzipping is true, we must fall back to reading in file
                cur_pos = fd.tell()
                fd.seek(0)  # test rewind ability
                fd.seek(0,2)  
                file_size = fd.tell() - cur_pos
                #print 'f %s %s' % (file_size, cur_pos)
                fd.seek(cur_pos)  # back to original position
                
            except (AssertionError, AttributeError, OSError), e: 
                # must abort when file not seekable, as re-try (in network.send_request) impossible
                raise ValueError('File postdata %s is not seekable. Cannot transfer' % k)
            
            try:
                name = fd.name.split('/')[-1]
                if isinstance(name, unicode):
                    name = name.encode('UTF-8')
            except AttributeError: #no name? revert to k 
                name = k

        buffer.write('--%s\r\n' % boundary)
        buffer.write('Content-Disposition: form-data; name="%s"; filename="%s"\r\n' \
                  % (k, name))
        buffer.write('Content-Type: %s\r\n' % get_content_type(name))
        buffer.write('Content-Length: %s\r\n' % file_size)        
        
        buffer.write('\r\n')

        if not contents:
            all_in_memory = False
        
        #print '[%s] transmit %s %s - in mem? %s contents? %s' % (sock, k,v, all_in_memory, bool(contents))
        if sock or not all_in_memory:
            assert not gzip #sanity check
            
            #reset buffer and update content-length if not returning result                
            buf_str = buffer.getvalue()            
            buffer.close()   
            cl += len(buf_str)
            buffer = StringIO() #regen buffer
            cl += file_size  
            
            if sock:             
                sock.sendall(buf_str)
                
                if contents:
                    sock.sendall(contents)
                else:
                    data=fd.read(CHUNK_SIZE)
                    try:
                        while data:
                            sock.sendall(data)
                            data=fd.read(CHUNK_SIZE)
                    finally:
                        if cur_pos != None: 
                            # rewind so retry will work
                            fd.seek(cur_pos)
            
        
        elif all_in_memory: #keep writing into RAM
            buffer.write(contents)
            
        buffer.write('\r\n')
            
    buffer.write('--%s--\r\n\r\n' % boundary)   
    
    if sock or not all_in_memory:
        buf_str = buffer.getvalue()
        if sock:
            sock.sendall(buf_str)        
        cl += len(buf_str)
        return None, cl        
    
    #in memory
    #When Using compressed content_len = compressed length (gzip
    # we need mod_wsgi middleware on the django server
    if gzip:
        buffer.close()
        buf_str = gzip_file.getvalue()
    else:        
        buf_str = buffer.getvalue()    
    
    return buf_str, len(buf_str)

def makeBodyFunction(v_vars, v_files, boundary):
    def theBody(httpcon):
        try:
            send_data(v_vars, v_files, boundary, gzip = False, sock = httpcon.sock)
        except socket.error, v:
            if v[0] == 32:      # Broken pipe
                httpcon.close()
            raise
    return theBody


class funcBodyHTTPConnection(httplib.HTTPConnection):
    """
    httplib connection that supports calling a function to handle body sending
    body takes a single argument -- this connection. 
    Note that automatic content-length calculation is not done if body is a function
    
    Also supports persistant
    """
    
    def _send_request(self, method, url, body, headers):
    
        if callable(body):
            httplib.HTTPConnection._send_request(self, method, url, None, headers)
            body(self)
        else:
            httplib.HTTPConnection._send_request(self, method, url, body, headers)
            
    @staticmethod
    def getresponse_static(self):
                
        """Get the response from the server.
        This is a static method used by getresponse in http and https"""
        
        #modified to read any extra data from pipe

        # if a prior response has been completed, then forget about it.
        if self._HTTPConnection__response:
            closed = self._HTTPConnection__response.isclosed()

            
            if not closed:
                self._HTTPConnection__response.read()
            
            self._HTTPConnection__response = None

        #
        # if a prior response exists, then it must be completed (otherwise, we
        # cannot read this response's header to determine the connection-close
        # behavior)
        #
        # note: if a prior response existed, but was connection-close, then the
        # socket and response were made independent of this HTTPConnection
        # object since a new request requires that we open a whole new
        # connection
        #
        # this means the prior response had one of two states:
        #   1) will_close: this connection was reset and the prior socket and
        #                  response operate independently
        #   2) persistent: the response was retained and we await its
        #                  isclosed() status to become true.
        #
        if self._HTTPConnection__state != httplib._CS_REQ_SENT or self._HTTPConnection__response:
            #print 'not ready %s %s' % (self._HTTPConnection__state, self._HTTPConnection__response)
            raise httplib.ResponseNotReady('State is %s' % self._HTTPConnection__state) 

        if self.debuglevel > 0:
            response = self.response_class(self.sock, self.debuglevel,
                                           strict=self.strict,
                                           method=self._method)
        else:
            response = self.response_class(self.sock, strict=self.strict,
                                           method=self._method)

        response.begin()
        assert response.will_close != httplib._UNKNOWN
        self._HTTPConnection__state = httplib._CS_IDLE

        if response.will_close:            
            # this effectively passes the connection to the response
            self.close()
            
            #hax:
            #self._HTTPConnection__response = response
            
        else:
            # remember this, so we can tell when it is complete
            self._HTTPConnection__response = response

        return response    
    
    def getresponse(self):
        return self.getresponse_static(self)        
        


# modified version from urllib2
class newHTTPAbstractHandler(urllib2.AbstractHTTPHandler):

    def __init__(self, debuglevel=0):
        urllib2.AbstractHTTPHandler.__init__(self, debuglevel)
        self.connections = {}
    
    def do_open(self, http_class, req):
        """Adds ability to reuse connections
        """
        host = req.get_host()
        if not host:
            raise urllib2.URLError('no host given')
        
        effective_host = getattr(req,'_tunnel_host', None) #check proxy first
        if not effective_host:
            effective_host = req.get_host()            
        #print 'effhost is %s. req host is %s' % (effective_host, host)
        
        conn_key = effective_host, http_class, thread.get_ident()  #good enough for scivm
        #print 'conn_key is ', conn_key
        h = self.connections.get(conn_key)
            
        if h:
            #print 'CONN: found reuse conn!', h.sock            
            reusing = True
            h._set_hostport(host,None)  #proxy tunneling renames this
        if not h:  #no cached conn - reconnect
            #print 'CONN: using a new connection'
            if hasattr(req,'timeout'):
                h = http_class(host, timeout=req.timeout) # will parse host:port
            else:
                h = http_class(host)  #py2.5
            h.set_debuglevel(self._debuglevel)
            self.connections[conn_key] = h
            reusing = False
        #To enable debugging
        #h.set_debuglevel(1)
        

        headers = dict(req.headers)
        headers.update(req.unredirected_hdrs)
        # We want to make an HTTP/1.1 request, but the addinfourl
        # class isn't prepared to deal with a persistent connection.
        # It will try to read all remaining data from the socket,
        # which will block while the server waits for the next request.
        # So make sure the connection gets closed after the (only)
        # request.
        
        close_connection = http_close_connection or headers.get("Connection") == "close" 
        
        if close_connection:
            headers["Connection"] = "close"
        
        headers = dict(
            (name.title(), val) for name, val in headers.items())

        if getattr(req, '_tunnel_host', None): #py2.5 lacks this
            tunnel_headers = {}
            proxy_auth_hdr = "Proxy-Authorization"
            if proxy_auth_hdr in headers:
                tunnel_headers[proxy_auth_hdr] = headers[proxy_auth_hdr]
                # Proxy-Authorization should not be sent to origin
                # server.
                del headers[proxy_auth_hdr]
            if hasattr(h, '_set_tunnel'):  #python2.6 and below
                h._set_tunnel(req._tunnel_host, headers=tunnel_headers)
            else:  #python2.7+ removed protection
                h.set_tunnel(req._tunnel_host, headers=tunnel_headers)

        for i in range(2): 
            #persistent connection handling via httplib2 
            try:
                #hack -- check if old response closed        
                #print 'start req'        
                h.request(req.get_method(), req.get_selector(), req.data, headers)
                #print 'end req'
            except socket.gaierror, err:
                h.close()
                raise urllib2.URLError(err)
            except (socket.error, httplib.HTTPException), e:
                #may still have response
                if isinstance(e, socket.error) and getattr(e, 'errno', e.args[0]) == errno.ECONNREFUSED:
                    h.close()
                    raise
                scicloudLog.info('Error raised by request. reinitiating connection. Exception was:', exc_info=1)
                h.close()
                if i == 0:
                    continue
                else:
                    raise
            except Exception: #cleanup on any request error
                h.close()
                raise 
                
            try:
                #print 'sock is', h.sock
                r = h.getresponse()
                #print 'post-end is', h.sock, r.will_close
            except Exception, err:  #known exceptions are (socket.error, httplib.HTTPException, AttributeError)
                #print 'socket closed!!', str(err)
                #socket closed -- retry?
                if i == 0 and reusing:
                #if False:
                    scicloudLog.info('Reconnecting socket due to:', exc_info=1) 
                    h.close()
                    h.connect()
                    continue
                    #raise  urllib2.URLError(err)
                else:
                    raise  urllib2.URLError(err)
                    #print err
                    #raise
            break


        # Pick apart the HTTPResponse object to get the addinfourl
        # object initialized properly.

        # Wrap the HTTPResponse object in socket's file object adapter
        # for Windows.  That adapter calls recv(), so delegate recv()
        # to read().  This weird wrapping allows the returned object to
        # have readline() and readlines() methods.

        # XXX It might be better to extract the read buffering code
        # out of socket._fileobject() and into a base class.

        r.recv = r.read
        try:
            fp = socket._fileobject(r, close=close_connection)  #set close to false to leave conn open
        except TypeError: #python pre-2.5.1 lacks close kwarg
            fp = socket._fileobject(r)

        resp = urllib2.addinfourl(fp, r.msg, req.get_full_url())
        resp.code = r.status
        resp.msg = r.reason
        
        return resp


    def do_request_(self, request):
        """Modified to support multipart data"""
        
        host = request.get_host()
        if not host:
            raise urllib2.URLError('no host given')

        if not request.has_data():
            request = urllib2.AbstractHTTPHandler.do_request_(self, request)
        else:
            data = request.get_data()
            #print 'data is %s' % data
            if callable(data):
                raise TypeError('data is a callable. This is not valid!')
            v_files=[]
            v_vars=[]
            # mapping object (dict)
            if type(data) == str:
                request = urllib2.AbstractHTTPHandler.do_request_(self, request)
            else:                         
                if hasattr(data, 'items'):
                    data = list(data.items())
                else:
                    if len(data) and not isinstance(data[0], tuple):
                        raise TypeError("not a valid non-string sequence or mapping object.")
                        
                    
                for (k, v) in data:
                    if isinstance(k, unicode):
                        k = k.encode('utf-8')
                        k = urllib.quote(k)
                    
                    if hasattr(v, 'read'):
                        v_files.append((k, v))
                    else:
                        if isinstance(v, unicode):
                            v = v.encode('utf8')
                        v_vars.append( (k, v) )
                        
                if len(v_files) == 0:
                    if v_vars:
                        data = urllib.urlencode(v_vars)
                        #print 'transmit %s' % data
                    else:
                        data = ""                      
                    if not request.has_header('Content-type'):
                        request.add_unredirected_header('Content-Type',
                                    'application/x-www-form-urlencoded')
                        request.add_unredirected_header('Content-length', '%d' % len(data))
                
                elif not request.has_header('Content-type') and len(v_files) > 0:
                        boundary = choose_boundary()
                        gzip_request = client_gzip and getattr(request,'use_gzip',False)
                        
                        data, content_len = send_data(v_vars, v_files, boundary, gzip_request)
                        request.add_unredirected_header('Content-Type',
                                    'multipart/form-data; boundary=%s' % boundary)
                        request.add_unredirected_header('Content-length', str(content_len))
                        
                        #gzip outbound requests:
                        if (gzip_request):
                            request.add_unredirected_header('Content-Encoding', 'gzip')
                
                #request's data is none if we need to do further evaluation
                if data is not None:
                    request.data = data
                else: #use callback
                    request.data = makeBodyFunction(v_vars, v_files, boundary)
                            
                #final steps (from urllib2.AbstractHTTPHandler)
                sel_host = host
                if hasattr(request,'has_proxy') and request.has_proxy():
                    scheme, sel = urllib.splittype(request.get_selector())
                    sel_host, sel_path = urllib.splithost(sel)
        
                if not request.has_header('Host'):
                    request.add_unredirected_header('Host', sel_host)
                for name, value in self.parent.addheaders:
                    name = name.capitalize()
                    if not request.has_header(name):
                        request.add_unredirected_header(name, value)              

        if (use_gzip):
            request.add_unredirected_header('Accept-Encoding', 'gzip')
            #pass

        return request
        
    def do_response_(self, request, response):         
        #print 'type', fp.__class__
        if response.headers and response.headers.getheader('Content-Encoding','') == 'gzip' and\
        getattr(request,'use_gzip',False):
                                        
            fp = GzipFile(filename=None,fileobj=StringIO(response.read()),mode='r')
            #fp = GzipFile(filename=None,fileobj=response,mode='r')
            #response needs to be StringIO(response.read()) without gzip streaming

            old_response = response
            response = urllib.addinfourl(fp, old_response.headers, old_response.url)
            if hasattr(old_response,'code'):
                response.code = old_response.code
            response.msg = old_response.msg
        return response

class newHTTPHandler(newHTTPAbstractHandler):
    http_request = newHTTPAbstractHandler.do_request_
    http_response = newHTTPAbstractHandler.do_response_
    
    def http_open(self, req):
        return self.do_open(funcBodyHTTPConnection, req)    

if hasattr(httplib, 'HTTPS'):
    class funcBodyHTTPSConnection(httplib.HTTPSConnection):
        """
        See docs for funcBodyHTTPConnection
        """
        
        def _send_request(self, method, url, body, headers):
            #print 'working req is %s\nbody=%s' % (headers, body)
            if callable(body):
                #print 'callable body'
                httplib.HTTPSConnection._send_request(self, method, url, None, headers)
                body(self)
            else:
                httplib.HTTPSConnection._send_request(self, method, url, body, headers)
                
        def getresponse(self):
            return funcBodyHTTPConnection.getresponse_static(self) #use this version
        
        def _send_output(self, message_body=None):
            """Fix Python 2.7 bug"""
            self._buffer.extend(("", ""))
            msg = "\r\n".join(self._buffer)
            del self._buffer[:]
            # If msg and message_body are sent in a single send() call,
            # it will avoid performance problems caused by the interaction
            # between delayed ack and the Nagle algorithim.
            
            if isinstance(message_body, str):
                try:
                    msg += message_body
                except UnicodeDecodeError: #non-ascii
                    pass
                else:
                    message_body = None
            self.send(msg)
            if message_body is not None:
                #message_body was not a string (i.e. it is a file) 
                self.send(message_body)


    
    class newHTTPSHandler(newHTTPAbstractHandler):
        https_request = newHTTPAbstractHandler.do_request_
        https_response = newHTTPAbstractHandler.do_response_
        
        def https_open(self, req):
            return self.do_open(funcBodyHTTPSConnection, req)    

#Input handling
_opener = None

#modified urlopen - supports python 2.5 and 2.6
def urlopen(url, data=None, timeout = None):
    global _opener
    if _opener is None:
        _opener = build_opener()
    if not timeout:
        if not hasattr(socket,'_GLOBAL_DEFAULT_TIMEOUT'):
            #python 2.5
            return _opener.open(url, data)
        else:
            timeout=socket._GLOBAL_DEFAULT_TIMEOUT
    if sys.version_info < (2,6):
        return _opener.open(url, data) # version 2.5 compat 
    else:
        return _opener.open(url, data, timeout)
        

#modified build_opener:
def build_opener(*handlers):
    """Create an opener object from a list of handlers.

    The opener will use several default handlers, including support
    for HTTP and FTP.

    If any of the handlers passed as arguments are subclasses of the
    default handlers, the default handlers will not be used.
    """
    import types
    def isclass(obj):
        return isinstance(obj, types.ClassType) or hasattr(obj, "__bases__")



    opener = urllib2.OpenerDirector()
    default_classes = [urllib2.ProxyHandler, urllib2.UnknownHandler, newHTTPHandler,
                       urllib2.HTTPDefaultErrorHandler, urllib2.HTTPRedirectHandler,
                       urllib2.FTPHandler, urllib2.FileHandler, urllib2.HTTPErrorProcessor]
    if hasattr(httplib, 'HTTPS'):
        default_classes.append(newHTTPSHandler)
    skip = set()

    if proxy_server:
        handlers = list(handlers)
        handlers.append(urllib2.ProxyHandler({'http': proxy_server,
                                                 'https': proxy_server}))    
    
    for klass in default_classes:
        for check in handlers:
            if isclass(check):
                if issubclass(check, klass):
                    skip.add(klass)
            elif isinstance(check, klass):
                skip.add(klass)
    for klass in skip:
        default_classes.remove(klass)

    for klass in default_classes:
        opener.add_handler(klass())
        
    for h in handlers:
        if isclass(h):
            h = h()
        opener.add_handler(h)
    return opener
    



"""
Binds scicloud object methods to outer modules

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from . import scicloudconfig as cc

#globals for scicloud.files
last_api_key = None
last_api_secretkey = None


def generate_scicloud(type='network', api_key=None, api_secretkey=None, server_url=None):
    """
    Generate a new Cloud object.  Generally setCloud should be used;
    This function should only be used if you wish to have duplicate sciclouds 
    running
    
    type is a string that specifies which scicloud to use.
    The two types are
    type='network' (The scicloud service provided by PiCloud. *Default*)
    type='local' (Simulated mode - Runs functions locally)
    
    If type is set to 'network', api_key and api_secretkey must be
    specified.  See your PiCloud account API Keys for this information
    """
    from .scicloud import Cloud
    
    if type == 'network':
        from .transport import HttpConnection
        from .transport import DependencyAdapter as CloudAdapter                                
        retval =  Cloud(CloudAdapter(HttpConnection(api_key, api_secretkey, server_url=server_url)))
        global last_api_key, last_api_secretkey
        last_api_key = api_key
        last_api_secretkey = api_secretkey
        return retval
    elif type == 'simulated':   
        try:
            from .transport.simulator import SimulatedConnection
        except ImportError:
            try:
                import multiprocessing
            except ImportError: #if mp error, fall back
                from .scicloudlog import scicloudLog
                scicloudLog.error('Multiprocessing is not installed. Simulator cannot be started. Falling back to regular PiCloud')
                return generate_scicloud()
            else:
                raise #something else has gone wrong    
        from .transport import SerializingAdapter                          
        return Cloud(SerializingAdapter(SimulatedConnection()))
    elif type == 'mp': #remove me
        from .transport import SerializingAdapter                                        
        #we check for mp existance in __init__ to allow module to be removed
        from .transport.local import MPConnection
        return Cloud(SerializingAdapter(MPConnection()))
    else:
        raise Exception('scicloud: Cloud generation type must be either network or local, not %s' % type)                            


def _setscicloud(module, type='network', api_key=None, api_secretkey=None, server_url=None, restart=False, immutable=False):
    """
    Set the scicloud that scicloud should use
    Binds module's scicloud to this scicloud
    
    type is a string that specifies which scicloud to use.
    The two types are:
    type='network' (Default: The scicloud service provided by PiCloud.)
    type='local' (Simulated mode - Runs functions locally)
    
    If type is set to 'network', api_key and api_secretkey must be
    specified.  See your PiCloud account API Keys for this information    
    
    restart is used internally to force the scicloud to be reconstructed and 
    reconnected
    
    If immutable is set, future setCloud calls will have no effect        
    """
    
    if not module.__immutable and (type != module.__type or restart or not module.__scicloud or 
       module.__scicloud.needs_restart(api_key=api_key,api_secretkey=api_secretkey, server_url=server_url)):
        
        # close any existing scicloud
        if module.__scicloud != None and module.__scicloud.opened:
            module.__scicloud.close()
                
        if type:
            scicloud = generate_scicloud(type, api_key, api_secretkey, server_url)        
            _bindscicloud(module,scicloud,type, immutable)
        else:
            module.__scicloud = None
            

def _bindscicloud(module, scicloud, type, immutable=False):
    module.__scicloud = scicloud 
    
    module.__type = type
    
    #bind every scicloud method listed in module's __all__ to module 
    for meth in module.__all__:
        if hasattr(scicloud,meth):
            setattr(module,meth,getattr(scicloud,meth))
    
    scicloud.parentModule = module.__name__
    
    if immutable:
        module.__immutable = True 
    
    
    


#!/usr/bin/env python

import os
import sys
import shutil

#fix a bug that setuptools creates:
import distutils.filelist
findall = distutils.filelist.findall  
from distutils.util import change_root, convert_path
from distutils.dir_util import mkpath
from distutils.file_util import copy_file

# For future use. Always will be true currently as we require setuptools
use_setuptools = False

def abort(msg):    
    print >> sys.stderr, msg
    sys.exit(1)


if sys.argv == ['setup.py', 'sdist']:
    print 'using regular setup tools'
    from distutils.core import setup
else:
    try:    
        from setuptools import setup
        distutils.filelist.findall = findall #reset
        use_setuptools = True
    except ImportError:
        abort('Distribute not installed. Cannot continue with installation. Get Distribute from http://pypi.python.org/pypi/distribute')

# bring in release_version
basedir = os.path.split(__file__)[0]

for line in open(os.path.join(basedir, 'src/versioninfo.py')).readlines():
    if (line.startswith('release_version')):
        exec(line.strip())
        

if sys.version < '2.6':
    abort('Python must be > 2.6 to use Science VM')
elif sys.version > '3':
    abort('This package cannot be used under Python3. Please get the Python3 package from http://www.scivm.com')


#Python2.5 lacks important packages like ssl and (simple)json. Discover if needed
requires = []
data_files = []
scripts = []
options = {}

#source install deps checking
try:
    import json
except ImportError:
    try:
        import simplejson
    except ImportError:
        if not use_setuptools:        
            abort('simplejson must be installed to use scivm\nDownload it at http://pypi.python.org/pypi/simplejson/')
        else:
            requires.append("simplejson")
            
try:
    import ssl
except ImportError:
    try:
        import OpenSSL
    except ImportError:
        if not use_setuptools:
            abort('OpenSSL python bindings not installed.  These are required to ensure a secure connection to Science VM.\nDownload them at http://pypi.python.org/pypi/ssl')
        else: #inject requirement
            requires.append('ssl')

#check for pywin32 on windows which is highly recommended:
#These can't be installed automatically, so just warn
if os.name == 'nt':
    try: 
        import win32con
    except ImportError:
        sys.stderr.write('Python for Windows extensions are highly recommended for using scivm.\nDownload them at http://sourceforge.net/projects/pywin32/')

if 'bdist_wininst' in sys.argv:
    # hot patch for bdist_wininst

    import distutils.command.bdist_wininst as bwin
    def get_exe_bytes (self):
        # Aaron's hack fix
        
        bv = 9.0
        if self.plat_name == 'win-amd64':
            sfix = '-amd64'
        else:
            sfix = ''
        directory = os.path.dirname(bwin.__file__)
        filename = os.path.join(directory, "wininst-%.1f%s.exe" % (bv, sfix))
        print 'using %s' % filename
        try:
            f = open(filename, "rb")
        except IOError, msg:
            from distutils.errors import DistutilsFileError
            raise DistutilsFileError, str(msg) + ', please install the python%s-dev package' % sys.version[:3]
        try:
            return f.read()
        finally:
            f.close()
            
    def get_installer_filename(self, fullname):
        # from http://bugs.python.org/issue8954
        
        # Factored out to allow overriding in subclasses

        # pure Python packages always have win32 extension
        # required for crossplatform behaviour
        if (not self.distribution.has_ext_modules() and 
            not self.distribution.has_c_libraries() and
            self.plat_name != 'win-amd64'):
            basename = "%s.%s" % (fullname, "win32")
        else:
            basename = "%s.%s" % (fullname, self.plat_name)

        if self.target_version:
            # if we create an installer for a specific python version,
            # it's better to include this in the name
            installer_name = os.path.join(self.dist_dir,
                                          "%s-py%s.exe" %
                                          (basename, self.target_version))
        else:
            installer_name = os.path.join(self.dist_dir, "%s.exe" % basename)
        return installer_name          
     
    bwin.bdist_wininst.get_exe_bytes = get_exe_bytes
    bwin.bdist_wininst.get_installer_filename = get_installer_filename
    
    if '--plat-name=win32' in sys.argv:
        distribute_file_path = 'extra/distribute-0.6.26.win32.msi'
    elif '--plat-name=win-amd64' in sys.argv:
        distribute_file_path = 'extra/distribute-0.6.26.win-amd64.msi'

    rsync_dir = 'extra/rsync'
    rsync_files = [os.path.join(rsync_dir, f) for f in os.listdir(rsync_dir)]
    #rsync_install_dir = os.path.join(sys.prefix, 'lib', 'site-packages',
    #                                 'sciscicloud', 'extras')

    data_files = [('scripts', [distribute_file_path]),
                  ('extras', rsync_files)]
    scripts = ['post_install.py']
    options = { 'bdist_wininst': {'install_script':'post_install.py'} }
    
    
    """hot fix for setuptools
    This script is for 32-bit so should always work"""
    from setuptools.command.easy_install import get_script_header, sys_executable, is_64bit, resource_string
    import re
       
    def get_script_args(dist, executable=sys_executable, wininst=False):
        """Yield write_script() argument tuples for a distribution's entrypoints"""
        spec = str(dist.as_requirement())
        header = get_script_header("", executable, wininst)
        for group in 'console_scripts', 'gui_scripts':
            for name, ep in dist.get_entry_map(group).items():
                script_text = (
                    "# EASY-INSTALL-ENTRY-SCRIPT: %(spec)r,%(group)r,%(name)r\n"
                    "__requires__ = %(spec)r\n"
                    "import sys\n"
                    "from pkg_resources import load_entry_point\n"
                    "\n"
                    "if __name__ == '__main__':"
                    "\n"
                    "    sys.exit(\n"
                    "        load_entry_point(%(spec)r, %(group)r, %(name)r)()\n"
                    "    )\n"
                ) % locals()
                if sys.platform=='win32' or wininst:
                    # On Windows/wininst, add a .py extension and an .exe launcher
                    if group=='gui_scripts':
                        ext, launcher = '-script.pyw', 'gui.exe'
                        old = ['.pyw']
                        new_header = re.sub('(?i)python.exe','pythonw.exe',header)
                    else:
                        ext, launcher = '-script.py', 'cli.exe'
                        old = ['.py','.pyc','.pyo']
                        new_header = re.sub('(?i)pythonw.exe','python.exe',header)
                    
                    if (sys.platform=='win32' and is_64bit()) or ('--plat-name=win-amd64' in sys.argv):
                        launcher = launcher.replace(".", "-64.")
                    else: # on linux (build system) always fall back to this
                        launcher = launcher.replace(".", "-32.")
                    if os.path.exists(new_header[2:-1]) or sys.platform!='win32':
                        hdr = new_header
                    else:
                        hdr = header
                    yield (name+ext, hdr+script_text, 't', [name+x for x in old])
                    yield (
                        name+'.exe', resource_string('setuptools', launcher),
                        'b' # write in binary mode
                    )
                else:
                    # On other platforms, we assume the right thing to do is to
                    # just write the stub with no extension.
                    yield (name, header+script_text)

    import setuptools.command.easy_install
    setuptools.command.easy_install.get_script_args = get_script_args

    

requires.append('setuptools')

dist = setup(
    name='sciscicloud',
    version=release_version,  #defined by versioninfo.py exec
    description='Science VM client-side library',      
    author='Science Automation, Inc.',
    author_email='contact@scivm.com',
    url='http://www.scivm.com',
    install_requires=requires,
    license='GNU LGPL',
    long_description=open('README.txt').read(),
    packages=['sciscicloud', 'sciscicloud.cli', 'sciscicloud.serialization', 'sciscicloud.transport', 'sciscicloud.util', 'sciscicloud.util.cloghandler', 'sciscicloud.shortcuts'],
    package_dir = {'sciscicloud': os.path.join(basedir, 'src')},
    package_data= {'sciscicloud.util.cloghandler' : ['README, PKG-INFO, LICENSE']}, 
    platforms=['CPython 2.6', 'CPython 2.7'],      
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        ],
    entry_points={'console_scripts': ['scivm = sciscicloud.cli.main:main',]},
    data_files = data_files,
    scripts = scripts,
    options = options
    )


if 'install' in sys.argv and '--single-version-externally-managed' not in sys.argv:
    if os.name != 'nt':
        
        prefix = os.path.normpath(sys.prefix)
        root = None
        try:
            installer = dist.command_obj['install']
            prefix = installer.prefix
            root = installer.root
        except Exception:
            pass        
        
        # Try to install manpage and bash scripts:
        extra_files = [
                 {'src' : 'bash_completion.d/scivm',
                 'dst' : '/etc/bash_completion.d/',
                 'name' : 'bash completion scripts'},
                 {'src' : 'doc/scivm.1',
                 'dst' : '%s/share/man/man1/' % prefix,
                 'name' : 'Science VM manpage'}
                 ]                         
        # copy code borrowed from distutils.command.install_data
        # We can't use original as it raises an exception if the file cant be written;
        # on error we just want to warn as these files aren't essential
        for file_info in extra_files:
            try:
                dst = convert_path(file_info['dst'])
                if root:
                    dst = change_root(root, dst)
                
                if root: 
                    # directories generally expected to exist already; if not we have wrong target directory
                    # for package maintainers using root, just build it to temporary directory
                    mkpath(dst)
                
                src = convert_path(file_info['src'])
                copy_file(src, dst)
            except (OSError, IOError, distutils.errors.DistutilsError), e:
                sys.stderr.write('Warning: Could not install %s to %s\n' % (file_info['name'], str(e)))                
        
    
    print ''
    print '********************************************************'
    print '********************************************************'
    print '***                                                  ***'
    print '***  Please run "scivm setup" to complete install  ***'
    if os.name == 'nt':
        print '*** scivm.exe is in the Scripts directory of Python***'            
    print '***                                                  ***'
    print '********************************************************'
    print '********************************************************'
    print ''



"""
PiCloud Simulator.
This is just a subclass of multiprocessing that modifies configuration options.
In the future, it may be more powerful

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from . import local
from .. import scicloudconfig as cc
from ..scicloudlog import scicloudLog

class SimulatedConnection(local.MPConnection):       
    simulatedForceSerializeDebugging = \
        cc.simulation_configurable('force_serialize_debugging',
                                  default=True,hidden=True,
                                  comment="If this is is set and scicloud is running in simulation mode, serialize_debugging  (see logging section) will be turned on")
    simulatedForceSerializeLogging = \
        cc.simulation_configurable('force_serialize_logging',
                                  default=True,
                                  comment="If this is is set and scicloud is running in simulation mode, serialize_logging  (see logging section) will be turned on")
    if simulatedForceSerializeLogging and not simulatedForceSerializeDebugging:
        simulatedForceSerializeDebugging = True
        scicloudLog.warning("force_serialize_logging implies force_serialize_debugging. Setting force_serialize_debugging to true")
    
    
    simulated_force_redirect_job_output = \
        cc.simulation_configurable('force_redirect_job_output', default=True, 
                                  comment="If set and scicloud is in simulation mode, forces redirect_job_output (see multiprocessing section) to true")
        
    def open(self):
        #modify mpconnection options here:
        if self.simulated_force_redirect_job_output:
            self.redirect_job_output = True
        
        if self.simulatedForceSerializeDebugging:
            self.adapter.serializeDebugging = True
        if self.simulatedForceSerializeLogging:
            self.adapter.serializeLogging = True
            
        self.adapter._configure_logging() #in case options changed
        
        local.MPConnection.open(self)
        
    def force_adapter_report(self):
        """
        Should the SerializationReport for the SerializationAdapter be coerced to be instantiated?
        """
        parent_force = local.MPConnection.force_adapter_report(self)
        return parent_force or self.simulatedForceSerializeLogging

           
    def connection_info(self):
        dict = local.MPConnection.connection_info(self)
        dict['connection_type'] = 'simulation'        
        return dict    

    def report_name(self):
        return 'Simulation'


"""
XRange serialization routines

Note: This code is in desperate need of cleanup. But options are limited.
Some bad API choices were made 2 years ago, and cannot be reverted

The code deals with several types.
The PiecewiseXrange is a list subclass that has xranges internally
Example:
[1,3,xrange(5,9), 20]
This defines the list [1,3,5,6,7,8,20] more compactly

This also deals with JSONable PiecewiseXrange. These encodings can then be encoded by JSON. They look like:
[1,3,['xrange', 5, 9, 1], 20]

Note:
Single integers are also considered PiecewiseXrange (e.g. 9 is a PiecewiseXrange)
Single xranges are also PiecewiseXrange. (e.g. xrange(10) is a PiecewiseXrange)


Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""


def xrange_params(xrangeobj):
    """Returns a 3 element tuple describing the xrange start, step, and len respectively
    
    Note: Only guarentees that elements of xrange are the same. parameters may be different.    
    e.g. xrange(1,1) is interpretted as xrange(0,0); both behave the same though w/ iteration
    """
    
    xrange_len = len(xrangeobj)
    if not xrange_len: #empty
        return (0,1,0) 
    start = xrangeobj[0]
    if xrange_len == 1: #one element
        return start, 1, 1 
    return (start, xrangeobj[1] - xrangeobj[0], xrange_len) 
    

"""Encoding (generally for placing within a JSON object):"""
    
def encode_maybe_xrange(obj, allowLists = True):
    """If obj is a picewiseXrange, make it a JSONable PiecewiseXrange
    Else return object
    If allowLists is False only an xrange can be encoded
    If true, PiecewiseXranges can be encoded 
    """
    if isinstance(obj, xrange):        
        return ['xrange'] + list(xrange_params(obj))
    if allowLists and isinstance(obj, list):
        return [encode_maybe_xrange(elm, allowLists = False) for elm in obj]
    return obj

def decode_maybe_xrange(obj, allowLists = True):
    """Decode an object that may be a JSONable piecewise xrange, into 
    a regular PiecewiseXrange
    Else return object"""
    if isinstance(obj, list):
        if len(obj) == 4 and obj[0] == 'xrange':  #an xrange object
            return xrange(obj[1], obj[1] + obj[2]*obj[3], obj[2])            
        elif allowLists: #decode internal xranges
            return [decode_maybe_xrange(elm, allowLists = False) for elm in obj]
    return obj

def decode_xrange_list(obj):
    """First call decode_maybe_xrange(obj)
    Then flatten the list, so no xranges are contained"""     
    
    obj =  decode_maybe_xrange(obj, allowLists = False) #outer layer might be an xrange
    if isinstance(obj, xrange):
        return list(obj) #convert to list
    outlst = []
    for elem in obj:
        if isinstance(elem, list) and len(elem) == 4 and elem[0] == 'xrange':
            outlst.extend(decode_maybe_xrange(elem, allowLists = False))
        else:
            outlst.append(elem)
    return outlst

"""Support for piece-wise xranges - effectively a list of xranges"""
class PiecewiseXrange(list):
    """A PiecewiseXrange is best explained with an example:
    
    [1,3,xrange(5,9), 20] is a PiecewiseXrange that defines    
    
    [1,3,5,6,7,8,20] more compactly
    
    This class exists to provide the my_iter method which iterates the elements defined.
    in example, myiter (in xrangeMode) would return:
    1,3,5,6,7,8,20 
    
    However, __iter__ returns:
    1,3,xrange(5,9), 20
    """
    
    def __init__(self):
        self._xrangeMode = False
        
    def to_xrange_mode(self):
        """Turn on xrange iterator"""
        self._xrangeMode = True
    
    def my_iter(self):
        """Ideally, this would overload __iter__, but that presents massive pickling issues"""
        if self._xrangeMode:
            return xrange_iter(self)
        else:
            return self.__iter__()

def xrange_iter(iterable):  #use a generator to iterate
    for elm in iterable.__iter__():
        if isinstance(elm, xrange):
            for subelm in elm:
                yield subelm
        else:
            yield elm

def maybe_xrange_iter(lst, coerce = True):
    """if lst is a PiecewiseXrange or coerce, iterate as though lst is a piecewise xrange
    otherwise, normal lst iteration 
    """
    if isinstance(lst, PiecewiseXrange):
        return lst.my_iter()
    elif coerce:
        return xrange_iter(lst)
    else:         
        return lst.__iter__()
        

def filter_xrange_list(func, xrange_list):
    """ Input is a PiecewiseXrange
        
        Returns a reasonably compact PiecewiseXrange consisting of all elements in the input where         
        where func(elem) evaluates to True
        
        Behavior notes:
        If an xrange reduces to one element, it is replaced by an integer
        e.g. xrange(n,n+1) = n
    """
    
    if not hasattr(xrange_list,'__iter__') or isinstance(xrange_list, xrange):
        xrange_list = [xrange_list]
    
    single_range = 2 #if > 0, then outList is just a single xrange
    no_xrange_output = True #outList has no xranges inserted
    
    outList = PiecewiseXrange()
    for elm in xrange_list:
        if isinstance(elm, (int, long)):  #elm is 
            if func(elm):
                outList.append(elm)
                single_range = 0 #individual elements present - so not single xrange
        elif isinstance(elm, xrange):            
            step = xrange_params(elm)[1]
            basenum = None
            for num in elm: #iterate through xrange
                if func(num):
                    if basenum == None:
                        basenum = num                        
                else:
                    if basenum != None: #push back xrange
                        if num - step == basenum: #only one element: push an integer
                            outList.append(basenum)
                            single_range = 0
                        else:
                            outList.append(xrange(basenum, num, step))
                            single_range-=1
                            no_xrange_output = False
                        basenum = None
            if basenum != None: #cleanup
                num+=step
                if num - step == basenum: #only one element: push an integer
                    outList.append(basenum)
                    single_range = 0
                else:
                    outList.append(xrange(basenum, num, step))
                    single_range-=1
                    no_xrange_output = False
        else:
            raise TypeError('%s (type %s) is not of type int, long or xrange' % (elm, type(elm)))
        
    if outList:
        if not no_xrange_output:
            if single_range > 0: #only one xrange appended - just return it
                return outList[0]
            else:
                outList.to_xrange_mode()
    return outList

def iterate_xrange_limit(pwx, limit):
    """    
    Split PiecewiseXrange *pwx* into individual pieceWiseXranges containing no more than limit elements 
    Iterator returns individual partitions
    """
    if not pwx: #empty list/non case
        return
    
    if isinstance(pwx, xrange):
        if len(pwx) <= limit:
            yield pwx
            return
        #needs to parition
        pwx = [pwx]
        
    #main algorithm
    outlist = []
    cnt = 0 #number of items appended
    for xr in pwx:
        if isinstance(xr, (int, long)):
            outlist.append(xr)
            cnt+=1
            if cnt >= limit:
                yield outlist
                outlist = []
                cnt =0
        elif isinstance(xr, xrange):
            while True:
                if len(xr) + cnt <= limit:
                    outlist.append(xr)
                    cnt+=len(xr)
                    break
                else: #break apart xrange
                    allowed = limit - cnt
                    start, step, xl_len = xrange_params(xr)
                    breakpoint = start+ step*allowed
                    to_app = xrange(start, breakpoint, step)
                    outlist.append(to_app)
                    yield outlist
                    outlist = []
                    cnt=0
                    xr = xrange(breakpoint,breakpoint+(xl_len-allowed)*step, step)
            if len(outlist) >= limit:
                yield outlist
                outlist = []  
                
        else:
            raise TypeError('%s (type %s) is not of type int, long or xrange' % (xr, type(xr)))
    if outlist:
        yield outlist    
    
"""
Quick unit test
(TODO: Move to unit tests)
"""

if __name__ == '__main__':
    
    for m in [xrange(0,10,2), xrange(0,1), xrange(0), xrange(0,-2), xrange(0,6), xrange(0,7,3)]:
        me = encode_maybe_xrange(m)
        n = decode_maybe_xrange(me)
        print m, me, n
        
        #split
        print list(iterate_xrange_limit(m,1000))
        print list(iterate_xrange_limit(m,3))
        print [list(x[0]) if type(x) is list else x for x in iterate_xrange_limit(m,3)]
    
    print 'filters!'
    xrl = filter_xrange_list(lambda x: x%10, xrange(10,100))
    print xrl
    
    print 'it limits!'
    lmt = list(iterate_xrange_limit(xrl,15))
    print 'chunks of 15', lmt
    
    lmt_read = [[list(x) for x in y] for y in lmt]
    lmt2 = [reduce(lambda x, y: x+y, x) for x in lmt_read] 
    print 'still 15', lmt2
    
    lenz = [len(x) for x in lmt2]
    print lenz
    
    print list(iterate_xrange_limit(xrange(25), 13))
    it = iterate_xrange_limit(xrange(25), 13)
    print 'iterate!'
    try:
        while True:
            v = it.next()
            print type(v), v
    except StopIteration:
        pass
        
        


"""
The package scicloud.shortcuts is a convenience package that holds functions and classes 
that couple other scicloud functions.

Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

def get_connection_info(jid, port, protocol='tcp', timeout=None):
    """Given a ``jid`` and a ``protocol`` ``port``, the job is listening on, return a dictionary containing:
    
        * address: IP address of server job is running on  
        * port: External (translated) port that can be used to connect to the listening ``port``        
    
    For certain services, the returned dictionary will include other relevant keys:
    * username: username to use to connect to a service running on ``port`` (available with ssh (port 22))
    
    This function blocks until job has opened the port for listening. An optional
    ``timeout`` may be specified in seconds.
    
    Example::
     
        >>> get_connection_info(jid,'8000')  # request the translation for jid's TCP Port 8000        
        {'address': 'ec2-23-22-210-36.compute-1.amazonaws.com',
         'port': '22000'}
     
    Note that this function is an alias scicloud.wait_for.port(jid, port, protocol, timeout) 
    """
    from .. import wait_for
    return wait_for.port(jid, port, protocol, timeout)
    



"""
ArgumentParsers for processing input to the PiCloud CLI.
"""
from __future__ import absolute_import
"""
Copyright (c) 2011 `PiCloud, Inc. <http://www.scivm.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import os
import errno

try:
    import argparse
except:
    from . import argparse


def shell_path_post_op(path):
    """Makes *path* specification in the CLI more natural for shell usage.
    
    Converts "." to "name"
    Converts "existing_dir/" to "existing_dir/name"
    
    Usage: put shell_path_post_op as a type when specifying a new argument
    for a parser"""
    
    def fix_path(**kwargs):
        
        # extract key_name 
        name = kwargs.get('name')
        if not name:
            name = kwargs.get('key_name')
        if not name:
            name = kwargs.get('obj_path')
        if not name:
            raise ValueError('no (key)name specified') 
        
        if os.path.isdir(path):
            return os.path.join(path, name)
        else: # dir may not exist yet
            dname = os.path.dirname(path)
            if dname:
                try:
                    os.makedirs(dname)
                except OSError, e:
                    if e.errno != errno.EEXIST:
                        raise
            basename = os.path.basename(path)
            if basename:  # dest is a file
                return path
            else:
                return os.path.join(path, name)
            
    return fix_path


def common_arg(*args, **kwargs):
    return args, kwargs


"""Primary Parser"""
scivm_parser = argparse.ArgumentParser(prog='scivm', description='PiCloud Command-Line Interface (CLI)')
scivm_parser.add_argument('--version', dest='_version', action='store_true', help='Print version')
scivm_parser.add_argument('-v', '--verbose', dest='_verbose', action='store_true', help='Increase amount of information you are given during command execution')
scivm_parser.add_argument('-o', '--output', dest='_output', default='default', choices=['json', 'no-header', 'default'], help='Format of output')
scivm_parser.add_argument('-a', '--api-key', dest='_api_key', type=int, help='API key to use')
scivm_parser.add_argument('-k', '--api-secretkey', dest='_api_secretkey', help='API secret key that matches the API key')
scivm_parser.add_argument('-s', '--simulate', dest='_simulate', action='store_true', help='Run command in simulator (see simulator documentation)')

scivm_subparsers = scivm_parser.add_subparsers(title='subcommands',
                                                   description='Subcommands and submodules of the PiCloud CLI. For detailed help, use "-h" after the subcommand/submodule. For example, for assistance with "exec", use "scivm exec -h".',
                                                   dest='_module',
                                                   )

"""Setup Parser"""
setup_parser = scivm_subparsers.add_parser('setup', description='Sets up the current machine to use PiCloud.',
                                             help='Set up the current machine to use PiCloud')
setup_parser.add_argument('--email', '-e', help='Email used to login to your PiCloud account')
setup_parser.add_argument('--password', '-p', help='Password used to login to your PiCloud account')
setup_parser.add_argument('--api-key', '-a', nargs='?', default=False, help='API Key to use. If specified without a value, a new API Key will be created without prompting')

"""Execute Parser"""
jid_help_str = "jid or comma-seperated list of jids. Ranges may be specified with '-', e.g. 1-4 for jobs 1,2, and 3"

# args common to call/map/rest-reg/etc
# We can't use parents as want these to appear AFTER data argument
common_args = [
    common_arg('-t', '--type', default='c1', choices=['c1', 'c2', 'f2', 'm1', 's1'], 
               help='Type of compute resource to use. Default c1'),
    common_arg('-r', '--return', metavar='FILENAME', dest='return_file',  
               help='If provided result of job will be contents of FILENAME. If not set, result is stdout'),
    common_arg('-e', '--env', help='Name of the environment to run job within'),
    common_arg('-v', '--vol', action='append', help='Provide job with access to this volume'),
    common_arg('-l', '--label', help='Label to attach to job. Allows for job filtering'),    
    common_arg('-c', '--cores', default=1, type=int, choices=[1, 2, 4, 8, 16], 
               help='Number of cores for your job to utilize'),
    common_arg('-i', '--ignore-exit-status', action='store_true',
               help='If provided, job will not error if command exits with a nonzero error code'),               
    common_arg('-w', '--cwd', metavar='DIRECTORY', help='Directory to execute command within'),
    common_arg('-m', '--max-runtime', type=int, help='Job will be terminated if it runs beyond integer MAX_RUNTIME minutes'),
    common_arg('-x', '--depends-on', metavar='JIDS', help='Jobs this depends on. JIDS are a %s' % jid_help_str),
    common_arg('--depends-on-errors', default='abort', choices=['abort', 'ignore'], 
               help='Policy for how a jid in depends-on erroring should be handled. If abort (default), set job to stalled. If ignore, treat error as satisfying'),
    common_arg('--not-restartable', dest='restartable', action='store_false',
               help='Indicates that job cannot be safely restarted on a hardware failure'),
    common_arg('--priority', type=int,
               help="Positive integer describing job's priority. PiCloud will try to run jobs with lower priorities earlier")

]               

exec_parser = scivm_subparsers.add_parser('exec', description='Executes a program on PiCloud through the shell. This is the shell version of scicloud.call',
                                            help="Execute a program on PiCloud through the shell")
exec_parser.add_argument('command', nargs=argparse.PARSER, help='Templated shell command to execute')
exec_parser.add_argument('-d', '--data', dest='args', metavar='PARAMETER=VALUE', action='append', 
                         help='Set template parameter to value')
for args, kwargs in common_args:
    exec_parser.add_argument(*args, **kwargs)

mapexec_parser = scivm_subparsers.add_parser('mapexec', description='Executes many programs in parallel on PiCloud through the shell. This is the shell version of scicloud.map. \
Number of programs will be determined by number of comma-seperated arguments provided to -n (--maparg)',
                                            help="Execute parallel programs on PiCloud through the shell")
mapexec_parser.add_argument('command', nargs=argparse.PARSER, help='Templated Shell command to execute.')
mapexec_parser.add_argument('-d', '--data', dest='args', metavar='PARAMETER=VALUE', action='append', 
                         help='For every mapjob, set template parameter to value')
mapexec_parser.add_argument('-n', '--map-data', dest='mapargs', metavar='PARAMETER=VALUE1,VALUE2,..', action='append',  
                         help="Specify values for template parameters. job1 parameter takes value1, job2 takes value2, etc.")
mapexec_parser.add_argument('-N', '--arg-file', dest='argfiles', metavar='PARAMETER=FILENAME', action='append',
                         help="Specify a file that specifies template parameter values. Each line corresponds to the argument for one job.")
mapexec_parser.add_argument('-k', '--copies', dest='duplicates', metavar='NUM_DUPLICATES', default=1, type=int,  
                        help='Duplicate jobs. Have NUM_DUPLICATES jobs take a given parameter value'),


for args, kwargs in common_args:
    mapexec_parser.add_argument(*args, **kwargs)


"""Additional PiCloud commands"""




jid_parser = argparse.ArgumentParser(add_help=False)
jid_parser.add_argument('jids', help=jid_help_str) 

status_parser = scivm_subparsers.add_parser('status', description='Obtain status of job(s)', 
                                              parents=[jid_parser], help='Status of job(s)')

result_parser = scivm_subparsers.add_parser('result', description='Obtain result of job(s)', 
                                              parents=[jid_parser], help='Result of job(s)')
result_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                           help='Error if job has not finished by this number of seconds')
result_parser.add_argument('-i', '--ignore-errors', action='store_true', default=False,
                           help='If job errored print Exception as return value and exit 0 rather than aborting with Exception')


join_parser = scivm_subparsers.add_parser('join', description='Wait until job(s) running on PiCloud are complete.', 
                                              parents=[jid_parser], help='Block until job(s) finished')
join_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                         help='Error if job has not finished by this number of seconds')

info_parser = scivm_subparsers.add_parser('info', description='Obtain information about job(s)', 
                                            parents=[jid_parser], help='Information about job(s)')
info_parser.add_argument('-o', '--output', dest='info_requested', default=None, 
                         help='comma seperated list of info desired. See docs for full listing. e.g. status, runtime')

kill_parser = scivm_subparsers.add_parser('kill', description='Abort running of job(s)',
                                            parents=[jid_parser], help='Kill job(s)')

delete_parser = scivm_subparsers.add_parser('delete', description='Delete job(s) from PiCloud', 
                                              parents=[jid_parser], help='Delete job(s)')

ssh_parser = scivm_subparsers.add_parser('ssh', description='Wait until job is processing and ssh into the container of a job',
                                           help='SSH into a job')
ssh_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                        help="Error if ssh not ready by this number of seconds.")
ssh_parser.add_argument('jid', type=int, help="job identifier")
ssh_parser.add_argument('cmd', nargs=argparse.REMAINDER, help='Optional command to run without opening interactive console (command should be usually be escaped with quotes)') 


ssh_info_parser = scivm_subparsers.add_parser('ssh-info', description='Obtain IP Address, Port, and username info to connect to a job for ssh/scp/etc. Use scivm ssh to log in to a job',
                                                help='Obtain SSH information')
ssh_info_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                             help="Error if ssh not not ready by this number of seconds.")
ssh_info_parser.add_argument('jid', type=int, help="job identifier") 


exec_shell_parser = scivm_subparsers.add_parser('exec-shell', description='Test your configuration by starting a session defined by your environment, volumes, etc. and logging into it',
                                            help="SSH into configuration")
exec_shell_parser.add_argument('--timeout', metavar='SECONDS', default=None, type=float,
                             help="Error if ssh not not ready by this number of seconds.")
exec_shell_parser.add_argument('-k','--keep-alive', action='store_true', help='Keep SSH job alive after all ssh connections terminated')
exec_shell_parser.add_argument('-t', '--type', default='c1', choices=['c1', 'c2', 'f2', 'm1', 's1'], 
               help='Type of compute resource to use. Default c1')
exec_shell_parser.add_argument('-e', '--env', help='Name of the environment to run job within')
exec_shell_parser.add_argument('-v', '--vol', action='append', help='Provide job with access to this volume')    
exec_shell_parser.add_argument('-c', '--cores', default=1, type=int, choices=[1, 2, 4, 8, 16], 
               help='Number of cores for your job to utilize')
exec_shell_parser.add_argument('-m', '--max-runtime', type=int, help='Job will be terminated if it runs beyond integer MAX_RUNTIME minutes')


"""Rest Parser"""
rest_parser = scivm_subparsers.add_parser('rest', description="Module for managing PiCloud REST interfaces", 
                                             help="Manage REST Interfaces")
rest_subparsers = rest_parser.add_subparsers(title='commands', dest='_command', help='command help')
rest_publish_parser = rest_subparsers.add_parser('publish', help='Publish a shell command to PiCloud which can be executed over REST')
rest_publish_parser.add_argument('label', help='Label to assign the published function')
rest_publish_parser.add_argument('command', nargs=argparse.PARSER, help='Templated shell command to execute')
for args, kwargs in common_args:
    dest = kwargs.get('dest')
    if dest in ['depends_on', 'depends_on_errors', 'label']: # not relevant to rest publish
        continue
    rest_publish_parser.add_argument(*args, **kwargs)

rest_list_parser = rest_subparsers.add_parser('list', help='List functions published to PiCloud')

rest_info_parser = rest_subparsers.add_parser('info', help='Retrieve information about a published function')
rest_info_parser.add_argument('label', help='Label of published function to get info about')

rest_invoke_parser = rest_subparsers.add_parser('invoke', help='Invoke a function published on PiCloud')
rest_invoke_parser.add_argument('label', help='Label of function to invoke')
rest_invoke_parser.add_argument('-d', '--data', dest='args', metavar='PARAMETER=VALUE', action='append', 
                         help='Set template parameter to value')

rest_mapinvoke_parser = rest_subparsers.add_parser('mapinvoke', help='Invoke a function published on PiCloud many times')
rest_mapinvoke_parser.add_argument('label', help='Label of function to invoke')
rest_mapinvoke_parser.add_argument('-d', '--data', dest='args', metavar='PARAMETER=VALUE', action='append', 
                         help='For every mapjob, set template parameter to value')
rest_mapinvoke_parser.add_argument('-n', '--map-data', dest='mapargs', metavar='PARAMETER=VALUE1,VALUE2,..', action='append',  
                         help="Specify template parameter for each value. e.g. job1 parameter takes value1, job2 takes value2, etc.")


rest_remove_parser = rest_subparsers.add_parser('remove', help='Remove a published function from PiCloud')
rest_remove_parser.add_argument('label', help='Label of published function to remove')


"""Files Parser"""
files_parser = scivm_subparsers.add_parser('files', description="Module for managing files stored on PiCloud's key-value store (Deprecated: see scivm bucket)", 
                                             help="(Deprecated: see bucket) Manage files on key-value store")
files_subparsers = files_parser.add_subparsers(title='commands', dest='_command', help='command help')

files_delete_parser = files_subparsers.add_parser('delete', help='Delete a file stored on PiCloud')
files_delete_parser.add_argument('name', default=None, help='Name of file stored on PiCloud')

files_get_parser = files_subparsers.add_parser('get', help='Retrieve a file from PiCloud')
files_get_parser.add_argument('name', help='Name of file in storage')
files_get_parser.add_argument('destination', type=shell_path_post_op, help='Local path to save file to')
#files_get_parser.add_argument('--start-byte', help='Starting byte')
#files_get_parser.add_argument('--end-byte', help='Ending byte')

files_getmd5_parser = files_subparsers.add_parser('get-md5', help='Get the md5 checksum of a file stored on PiCloud')
files_getmd5_parser.add_argument('name', default=None, help='Name of file stored on PiCloud')

files_list_parser = files_subparsers.add_parser('list', help='List files in PiCloud Storage')

files_put_parser = files_subparsers.add_parser('put', help='Store a file on PiCloud')
files_put_parser.add_argument('source', help='Local path to file')
files_put_parser.add_argument('name', default=None, help='Name for file to be stored on PiCloud')

files_syncfromscicloud_parser = files_subparsers.add_parser('sync-from-scicloud', help='Download file if it does not exist locally or has changed')
files_syncfromscicloud_parser.add_argument('name', default=None, help='Name of file stored on PiCloud')
files_syncfromscicloud_parser.add_argument('destination', type=shell_path_post_op, help='Local path to save file to')

files_synctoscicloud_parser = files_subparsers.add_parser('sync-to-scicloud', help='Upload file only if it does not exist on PiCloud or has changed')
files_synctoscicloud_parser.add_argument('source', default=None, help='local path to file')
files_synctoscicloud_parser.add_argument('name', default=None, help='Name for file to be stored on PiCloud')

"""Bucket Parser"""
bucket_parser = scivm_subparsers.add_parser('bucket', description="Module for managing bucket objects stored on PiCloud's key-value store",
                                              help="Manage data objects in your bucket")
bucket_subparsers = bucket_parser.add_subparsers(title='commands', dest='_command', help='command help')

bucket_remove_parser = bucket_subparsers.add_parser('remove', help='Remove a bucket object from PiCloud')
bucket_remove_parser.add_argument('obj_paths', metavar='obj-path', default=None, help='Name of bucket object')
bucket_remove_parser.add_argument('-p', '--prefix', default=None, type=str, help='If provided, prepend PREFIX/ to obj-path')

bucket_remove_prefix_parser = bucket_subparsers.add_parser('remove-prefix', help='Remove list of bucket objects from PiCloud beginnign with prefix')
bucket_remove_prefix_parser.add_argument('prefix', default=None, help='Remove all objects beginning witht his prefix')


bucket_get_parser = bucket_subparsers.add_parser('get', help='Retrieve a bucket object from PiCloud')
bucket_get_parser.add_argument('obj_path', metavar='obj-path', help='Name of bucket object')
bucket_get_parser.add_argument('file_path', metavar='file-path', type=shell_path_post_op, help='Local path to save bucket object to')
bucket_get_parser.add_argument('-p', '--prefix', default=None, type=str, help='If provided, prepend PREFIX/ to obj-path')
bucket_get_parser.add_argument('--start-byte', default=0, type=int, help='Starting byte')
bucket_get_parser.add_argument('--end-byte', default=None, type=int,help='Ending byte')


bucket_getmd5_parser = bucket_subparsers.add_parser('get-md5', help='Get the md5 checksum of a bucket object stored on PiCloud')
bucket_getmd5_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Name of bucket object')
bucket_getmd5_parser.add_argument('-p', '--prefix', default=None, type=str, help='If provided, prepend PREFIX/ to obj-path')

bucket_iterlist_parser = bucket_subparsers.add_parser('iterlist', help='List bucket object keys stored on PiCloud (complete listing)')
bucket_iterlist_parser.add_argument('-f', '--folderize', action="store_true", default=False,  
                                    help='Treat listing as directory based; compact keys containing "/" into a single folder')
bucket_iterlist_parser.add_argument('-p', '--prefix', default=None, type=str, 
                                    help='Return only keys beginning with PREFIX. If folderize, list a folder by setting PREFIX to the folder')

bucket_list_parser = bucket_subparsers.add_parser('list', help='List bucket object keys stored on PiCloud (result may be incomplete)')
bucket_list_parser.add_argument('-f', '--folderize', action="store_true", default=False,  
                                help='Treat listing as directory based; compact keys containing "/" into a single folder')
bucket_list_parser.add_argument('-p', '--prefix', default=None, type=str, 
                                help='Return only keys beginning with PREFIX. If folderize, list a folder by setting PREFIX to the folder')
bucket_list_parser.add_argument('-m', '--marker', default=None, type=str, 
                                help='Return only keys where key > marker')
bucket_list_parser.add_argument('-k', '--max-keys', default=1000, type=int, 
                                help='Maxmimum number of names that can be returned (max 1000)')

bucket_info_parser = bucket_subparsers.add_parser('info', help='Return information about bucket object')
bucket_info_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Name of bucket object')
bucket_info_parser.add_argument('-p', '--prefix', default=None, type=str,
                                help='If provided, prepend PREFIX/ to obj-path')

bucket_put_parser = bucket_subparsers.add_parser('put', help='Store a bucket object on PiCloud')
bucket_put_parser.add_argument('file_path', metavar='file-path', help='Local path to upload from')
bucket_put_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Key (name) of the bucket object to be stored on PiCloud')
bucket_put_parser.add_argument('-p', '--prefix', default=None, type=str,
                               help='If provided, prepend PREFIX/ to obj-path')

bucket_syncfromscicloud_parser = bucket_subparsers.add_parser('sync-from-scicloud', help='Download bucket object only if it has changed')
bucket_syncfromscicloud_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Name of bucket object in storage')
bucket_syncfromscicloud_parser.add_argument('file_path', metavar='file-path', type=shell_path_post_op, help='Local path to save bucket object to')
bucket_syncfromscicloud_parser.add_argument('-p', '--prefix', default=None, type=str,
                                         help='If provided, prepend PREFIX/ to obj-path')


bucket_synctoscicloud_parser = bucket_subparsers.add_parser('sync-to-scicloud', help='Upload bucket object only if it has changed')
bucket_synctoscicloud_parser.add_argument('file_path', metavar='file-path', default=None, help='Local path to upload from')
bucket_synctoscicloud_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Name for bucket object to be stored on PiCloud')
bucket_synctoscicloud_parser.add_argument('-p', '--prefix', default=None, type=str,
                                       help='If provided, prepend PREFIX/ to obj-path')


bucket_make_public_parser = bucket_subparsers.add_parser('make-public', help='Makes the bucket object publicly accessible by a URL')
bucket_make_public_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Name of bucket object in storage')
bucket_make_public_parser.add_argument('-p', '--prefix', default=None, type=str,
                                       help='If provided, prepend PREFIX/ to obj-path')
bucket_make_public_parser.add_argument('-d', '--header', dest='header_args', metavar='PARAMETER=VALUE', action='append', 
                                       help='Customize headers that should in HTTP response. See Python docs on scicloud.bucket.make_public')
bucket_make_public_parser.add_argument('-r', '--reset-headers', action='store_true', default=False,
                                       help='Clear all custom HTTP Headers')

bucket_make_private_parser = bucket_subparsers.add_parser('make-private', help='Removes the publicly accessible URL associated with the bucket object')
bucket_make_private_parser.add_argument('obj_path', metavar='obj-path', default=None, help='Name of bucket object in storage')
bucket_make_private_parser.add_argument('-p', '--prefix', default=None, type=str,
                                        help='If provided, prepend PREFIX/ to obj-path')

bucket_is_public = bucket_subparsers.add_parser('is-public', help='Determine if bucket object is publicly accessible by a URL')
bucket_is_public.add_argument('obj_path', metavar='obj-path', default=None, help='Name of bucket object in storage')
bucket_is_public.add_argument('-p', '--prefix', default=None, type=str,
                              help='If provided, prepend PREFIX/ to obj-path')

bucket_public_url_folder = bucket_subparsers.add_parser('public-url-folder', help='Return HTTP path that begins all your public bucket URLs')

bucket_mpsafe_get_parser = bucket_subparsers.add_parser('mpsafe-get', help='Atomically retrieve a bucket object from PiCloud. Use this in lieu of get/sync when multiple processes need to get a bucket')
bucket_mpsafe_get_parser.add_argument('obj_path', metavar='obj-path', help='Name of bucket object')
bucket_mpsafe_get_parser.add_argument('file_path', metavar='file-path', type=shell_path_post_op, help='Local path to save bucket object to')
bucket_mpsafe_get_parser.add_argument('-p', '--prefix', default=None, type=str, help='If provided, prepend PREFIX/ to obj-path')
bucket_mpsafe_get_parser.add_argument('--start-byte', default=0, type=int, help='Starting byte')
bucket_mpsafe_get_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                           help="Error if timeout not acquired by by this number of seconds.")
bucket_mpsafe_get_parser.add_argument('-s', '--sync', dest='do_sync', action='store_true',
                               help='If set, obtain file if local is different from version on PiCloud; if not set, only obtain file it does not exist locally')


"""Realtime Parser"""
realtime_parser = scivm_subparsers.add_parser('realtime', description='Module for managing realtime cores.', help = "Manage realtime cores")
realtime_subparsers = realtime_parser.add_subparsers(title='commands', dest='_command', help='command help')

realtime_list_parser = realtime_subparsers.add_parser('list', help='List realtime reservations')

realtime_release_parser = realtime_subparsers.add_parser('release', help='Release realtime cores')
realtime_release_parser.add_argument('request-id')

realtime_request_parser = realtime_subparsers.add_parser('request', help='Request realtime cores')
realtime_request_parser.add_argument('-m', '--max-duration', metavar='HOURS', default=None, type=int,
                                     help='Time for request to be active.')
realtime_request_parser.add_argument('type', choices=['c1', 'c2', 'f2', 'm1', 's1'], help='The type of core to reserve.')
realtime_request_parser.add_argument('cores', type=int, help='The number of cores to reserve.')

""" Volume Parser """
volume_parser = scivm_subparsers.add_parser('volume', description='Module for managing volumes stored on PiCloud', help = "Manage volumes")
volume_subparsers = volume_parser.add_subparsers(title='commands', dest='_command', help='command help')

volume_list_parser = volume_subparsers.add_parser('list', help='List existing volumes', description='Lists existing scicloud volumes')
volume_list_parser.add_argument('-n', '--name', nargs='+', help='Name(s) of volume to list')
volume_list_parser.add_argument('-d', '--desc', action='store_true', default=False, help='Print volume description')

volume_create_parser = volume_subparsers.add_parser('create', help='Create a volume on PiCloud')
volume_create_parser.add_argument('name', help='Name of the volume to create (max 64 chars)')
volume_create_parser.add_argument('mount_path', metavar='mount-path', help='Mount point (is relative then relative to /home/scivm) where jobs should expect this volume.')
volume_create_parser.add_argument('-d', '--desc', default=None, help='Description of the volume (max 1024 chars)')

volume_mkdir_parser = volume_subparsers.add_parser('mkdir', help='Create directory(ies) at scicloud volume [path]')
volume_mkdir_parser.add_argument('volume_path', metavar='volume-path', nargs='+', help='Cloud volume path where directory should be created')
volume_mkdir_parser.add_argument('-p', '--parents', action='store_true', default=False, help='Make necessary parents')

volume_sync_parser = volume_subparsers.add_parser('sync', help='Sync local directory to volume on PiCloud', description='Syncs a local path and a scicloud volume.', formatter_class=argparse.RawDescriptionHelpFormatter,)
volume_sync_parser.add_argument('source', nargs='+', help='Source path that should be synced')
volume_sync_parser.add_argument('dest', help='Destination path that should be synced')
volume_sync_parser.add_argument('-d', '--delete', action='store_true', help='Delete destination files that do not exist in source')

volume_delete_parser = volume_subparsers.add_parser('delete', help='Delete a scicloud volume')
volume_delete_parser.add_argument('name', help='Name of the scicloud volume to delete')

volume_ls_parser = volume_subparsers.add_parser('ls', help='List the contents of a scicloud volume [path]')
volume_ls_parser.add_argument('volume_path', metavar='volume-path', nargs='+', help='Cloud volume path whose contents should be shown')
volume_ls_parser.add_argument('-l', '--extended-info', action='store_true', default=False, help='Use long listing format')

volume_rm_parser = volume_subparsers.add_parser('rm', help='Remove contents from a scicloud volume')
volume_rm_parser.add_argument('volume_path', metavar='volume-path', nargs='+', help='Cloud volume path whose contents should be removed')
volume_rm_parser.add_argument('-r', '--recursive', action='store_true', default=False, help='Remove directories and their contents recursively')

""" Environment Parser """
environment_parser = scivm_subparsers.add_parser('env', description='Module for managing custom environments', help = "Manage environments")
environment_subparsers = environment_parser.add_subparsers(title='commands', dest='_command', help='command help')

environment_list_parser = environment_subparsers.add_parser('list', help='List existing environments', description='Lists existing custom environments')
environment_list_parser.add_argument('-n', '--name', nargs='+', help='Name(s) of environments to list')

environment_list_bases_parser = environment_subparsers.add_parser('list-bases', help='List base environments', description='Lists base environments from which to build custom environments')

environment_create_parser = environment_subparsers.add_parser('create', help='Create a new custom environment')
environment_create_parser.add_argument('name', help='Name of the environment to create (max 30 chars)')
environment_create_parser.add_argument('base', help='Name of the base OS to use (use "env list_bases" for base names)')
environment_create_parser.add_argument('-d', '--desc', default=None, help='Description of the environment (max 2000 chars)')

environment_clone_parser = environment_subparsers.add_parser('clone', help='Clone an existing custom environment')
environment_clone_parser.add_argument('parent_name', metavar='parent-name', help='Name of the environment to clone')
environment_clone_parser.add_argument('new_name', metavar='new-name', nargs='?', default=None, help='Name of cloned environment (default adds _clone to parent name)')
environment_clone_parser.add_argument('-d', '--new-desc', default=None, help='Description of the environment (max 2000 chars)')

environment_modify_parser = environment_subparsers.add_parser('modify', help='Modify and existing environment')
environment_modify_parser.add_argument('name', help='Name of existing environment to modify')

environment_get_hostname_parser = environment_subparsers.add_parser('get-hostname', help='Get the hostname of environment setup server')
environment_get_hostname_parser.add_argument('name', help='Name of the environment to query for setup server hostname')

environment_get_keypath_parser = environment_subparsers.add_parser('get-keypath', help='Get the path of the ssh key for setup servers')

environment_save_parser = environment_subparsers.add_parser('save', help='Save the current modifications for an environment without shutting down setup server')
environment_save_parser.add_argument('name', help='Name of the environment to save')

environment_shutdown_parser = environment_subparsers.add_parser('shutdown', help='Shutdown the setup server for an environment without saving changes')
environment_shutdown_parser.add_argument('name', help='Name of the environment to shutdown')

environment_save_shutdown_parser = environment_subparsers.add_parser('save-shutdown', help='Save the current modifications for an environment and shutdown the setup server')
environment_save_shutdown_parser.add_argument('name', help='Name of the environment to save and shutdown')

environment_delete_parser = environment_subparsers.add_parser('delete', help='Delete an environment')
environment_delete_parser.add_argument('name', help='Name of the environment to delete')

environment_edit_info_parser = environment_subparsers.add_parser('edit-info', help='Edit name and description of an environment')
environment_edit_info_parser.add_argument('name', help='Name of the environment to edit.')
environment_edit_info_parser.add_argument('-n', '--new-name', default=None, help='New name of the environment (max 30 chars)')
environment_edit_info_parser.add_argument('-d', '--new-desc', default=None, help='New description of the environment (max 2000 chars)')

environment_ssh_parser = environment_subparsers.add_parser('ssh', help='Connect to an environment via ssh')
environment_ssh_parser.add_argument('name', help='Name of the environment to connect')
environment_ssh_parser.add_argument('cmd', nargs=argparse.REMAINDER, help='Optional command to run without opening interactive console (command should be usually be escaped with quotes)')

environment_rsync_parser = environment_subparsers.add_parser('rsync', help='rsync to environment setup server')
environment_rsync_parser.add_argument('src_path', metavar='src-path', nargs='+', help='Source path to rsync')
environment_rsync_parser.add_argument('dest_path', metavar='dest-path', help='Destination path to rsync')
environment_rsync_parser.add_argument('-d', '--delete', action='store_true', help='Delete destination files that do not exist in source')

environment_run_script_parser = environment_subparsers.add_parser('run-script', help='Run a script on the environment setup server')
environment_run_script_parser.add_argument('name', help='Name of the environment to run the script')
environment_run_script_parser.add_argument('filename', help='Local path to script to run on the environment setup server')


""" Queue Parser """

"""
queue_parser = scivm_subparsers.add_parser('queue', description="Module for managing PiCloud Queues", 
                                             help="Manage queues")
queue_parsers = queue_parser.add_subparsers(title='commands', dest='_command', help='command help')

queue_attach_parser = queue_parsers.add_parser('attach', help='Attach a command to process queue messages')
queue_attach_parser.add_argument('name', help='Name of queue')
queue_attach_parser.add_argument('messager_handler', metavar='message-handler', help='Crontab schedule. See documentation for format details')
queue_attach_parser.add_argument('command', nargs=argparse.PARSER, help='Templated shell command to execute')
for args, kwargs in common_args:
    dest = kwargs.get('dest')
    if dest in ['depends_on', 'depends_on_errors', 'label']: # not relevant to rest publish
        continue
    queue_attach_parser.add_argument(*args, **kwargs)

queue_list_parser = queue_parsers.add_parser('list', help='List queues')
queue_create_parser = queue_parsers.add_parser('create', help='Create queue')
queue_create_parser.add_argument('name', help='Name of new queue')
queue_delete_parser = queue_parsers.add_parser('delete', help='Delete queue')
queue_delete_parser.add_argument('name', help='Name of queue to delete')
"""

""" Cron Parser """
cron_parser = scivm_subparsers.add_parser('cron', description="Module for managing PiCloud Crons, periodically invoked functions on PiCloud", 
                                             help="Manage periodically invoked functions")
cron_parsers = cron_parser.add_subparsers(title='commands', dest='_command', help='command help')
cron_register_parser = cron_parsers.add_parser('register', help='Register a function which will be periodically invoked on PiCloud according to a cron schedule')
cron_register_parser.add_argument('label', help='Label to assign the registered cron')
cron_register_parser.add_argument('schedule', metavar='CRONTAB', help='Crontab schedule. See documentation for format details')
cron_register_parser.add_argument('command', nargs=argparse.PARSER, help='Shell command to execute')
for args, kwargs in common_args:
    dest = kwargs.get('dest')
    if dest in ['depends_on', 'depends_on_errors', 'label']: # not relevant to rest publish
        continue
    cron_register_parser.add_argument(*args, **kwargs)

cron_list_parser = cron_parsers.add_parser('list', help='List crons registered on PiCloud')
cron_info_parser = cron_parsers.add_parser('info', help='Retrieve information about a registered cron')
cron_info_parser.add_argument('label', help='Label of registered cron to get info about')

cron_deregister_parser = cron_parsers.add_parser('deregister', help='Deregister (delete) the cron specified by label')
cron_deregister_parser.add_argument('label', help='Label of registered cron to deregister')

cron_manualrun_parser = cron_parsers.add_parser('run', help='Manually run the command stored with a cron, ignoring schedule')
cron_manualrun_parser.add_argument('label', help="Label of registered cron's function to manually invoke")

"""Wait for"""
wait_for_parser = scivm_subparsers.add_parser('wait-for', description="Module for waiting on various aspects of a job. See subcommands", 
                                             help="Wait on aspects of a job")
wait_for_parsers = wait_for_parser.add_subparsers(title='commands', dest='_command', help='command help')
port_parser = wait_for_parsers.add_parser('port', description='Wait for job to open port for listening. Return translated external IP address and port', 
                                              help="Wait for port to open; and translate it")
port_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                           help="Error if port is not ready by this number of seconds.")
port_parser.add_argument('-p', '--protocol', default='tcp', type=str, choices=['tcp', 'udp'],
                           help="Protocol port is listening on. Default is tcp")
port_parser.add_argument('jid', type=int, help="job identifier") 
port_parser.add_argument('port', type=int, help="Port to translate")

status_parser = wait_for_parsers.add_parser('status', description='Wait for job to have a a status', 
                                              help="Wait for status")
status_parser.add_argument('-t', '--timeout', metavar='SECONDS', default=None, type=float,
                           help="Error if port is not ready by this number of seconds.")
status_parser.add_argument('jid', type=int, help="job identifier")
from ..wait_for import _status_transitions 
status_parser.add_argument('test_status', metavar='status', type=str, choices=_status_transitions.keys(),
                           help='status to wait for')



"""
All piscicloud transport related functions

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""


from .network import HttpConnection
from .codedependency import DependencyManager
from .adapter import SerializingAdapter, DependencyAdapter